# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.utils.py3 import httplib
from libcloud.common.types import LibcloudError
from libcloud.compute.base import (
    Node,
    NodeSize,
    NodeImage,
    NodeLocation,
    StorageVolume,
    VolumeSnapshot,
    NodeAuthPassword,
)
from libcloud.test.secrets import ECS_PARAMS
from libcloud.compute.types import NodeState, StorageVolumeState
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.ecs import ECSDriver


class ECSDriverTestCase(LibcloudTestCase):
    region = "cn-qingdao"
    zone = "cn-qingdao-b"
    image_id = "ubuntu1404_64_20G_aliaegis_20150325.vhd"

    def setUp(self):
        ECSMockHttp.test = self
        ECSDriver.connectionCls.conn_class = ECSMockHttp
        ECSMockHttp.use_param = "Action"
        ECSMockHttp.type = None

        self.driver = ECSDriver(*ECS_PARAMS, region=self.region)
        self.fake_size = NodeSize(
            "ecs.t1.small", "ecs t1 small", None, None, None, None, self.driver
        )
        self.fake_image = NodeImage(self.image_id, name="ubuntu 14.04 64bit", driver=self.driver)
        self.fake_node = Node(
            id="fake-node1",
            name="fake-node",
            state=NodeState.RUNNING,
            public_ips=None,
            private_ips=None,
            driver=self.driver,
        )
        self.fake_volume = StorageVolume(
            id="fake-volume1",
            name="fake-volume",
            size=self.fake_size,
            driver=self.driver,
        )
        self.fake_snapshot = VolumeSnapshot(id="fake-snapshot1", driver=self.driver)
        self.fake_location = NodeLocation(
            id=self.region, name=self.region, country=None, driver=self.driver
        )
        self.fake_instance_id = "fake_instance_id"
        self.fake_security_group_id = "fake_security_group_id"

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        self.assertIsNotNone(nodes)
        self.assertEqual(1, len(nodes))
        node = nodes[0]
        self.assertEqual("iZ28n7dkvovZ", node.name)
        self.assertEqual("i-28n7dkvov", node.id)
        self.assertEqual(NodeState.PENDING, node.state)
        self.assertEqual(1, len(node.public_ips))
        self.assertEqual("114.215.124.73", node.public_ips[0])
        self.assertEqual(1, len(node.private_ips))
        self.assertEqual("10.163.197.74", node.private_ips[0])
        expected_extra = {
            "image_id": "ubuntu1404_64_20G_aliaegis_20150325.vhd",
            "description": "",
            "instance_type_family": "ecs.t1",
            "zone_id": "cn-qingdao-b",
            "internet_charge_type": "PayByTraffic",
            "serial_number": "ca0122d9-374d-4fce-9fc0-71f7c3eaf1c3",
            "io_optimized": "false",
            "device_available": "true",
            "instance_network_type": "classic",
            "hostname": "iZ28n7dkvovZ",
            "instance_type": "ecs.t1.small",
            "creation_time": "2015-12-27T07:35Z",
            "instance_charge_type": "PostPaid",
            "expired_time": "2999-09-08T16:00Z",
        }
        self._validate_extras(expected_extra, node.extra)
        vpc = {
            "vpc_id": "",
            "vswitch_id": "",
            "private_ip_address": None,
            "nat_ip_address": "",
        }
        self._validate_extras(vpc, node.extra["vpc_attributes"])
        eip_address = {
            "allocation_id": "",
            "ip_address": "",
            "internet_charge_type": "",
            "bandwidth": None,
        }
        self._validate_extras(eip_address, node.extra["eip_address"])
        self.assertIsNone(node.extra["operation_locks"]["lock_reason"])

    def test_list_nodes_with_ex_node_ids(self):
        ECSMockHttp.type = "list_nodes_ex_node_ids"
        nodes = self.driver.list_nodes(ex_node_ids=["i-28n7dkvov", "not-existed-id"])
        self.assertIsNotNone(nodes)

    def test_list_nodes_with_ex_filters(self):
        ECSMockHttp.type = "list_nodes_ex_filters"
        nodes = self.driver.list_nodes(ex_filters={"ZoneId": self.zone})
        self.assertIsNotNone(nodes)

    def _validate_extras(self, expected, actual):
        self.assertIsNotNone(actual)
        for key, value in iter(expected.items()):
            self.assertTrue(key in actual)
            self.assertEqual(
                value,
                actual[key],
                (
                    "extra %(key)s not equal, "
                    'expected: "%(expected)s", '
                    'actual: "%(actual)s"' % {"key": key, "expected": value, "actual": actual[key]}
                ),
            )

    def test_create_node(self):
        ECSMockHttp.type = "create_node"
        name = "test_create_node"
        node = self.driver.create_node(
            name=name,
            image=self.fake_image,
            size=self.fake_size,
            ex_security_group_id="sg-28ou0f3xa",
            ex_description="description",
            ex_internet_charge_type="PayByTraffic",
            ex_internet_max_bandwidth_out=1,
            ex_internet_max_bandwidth_in=200,
            ex_hostname="hostname",
            auth=NodeAuthPassword("password"),
            ex_io_optimized=True,
            ex_system_disk={
                "category": "cloud",
                "disk_name": "root",
                "description": "sys",
            },
            ex_vswitch_id="vswitch-id1",
            ex_private_ip_address="1.1.1.2",
            ex_client_token="client_token",
        )
        self.assertIsNotNone(node)

    def test_create_node_with_data_disk(self):
        ECSMockHttp.type = "create_node_with_data"
        self.name = "test_create_node"
        self.data_disk = {
            "size": 5,
            "category": self.driver.disk_categories.CLOUD,
            "disk_name": "data1",
            "description": "description",
            "device": "/dev/xvdb",
            "delete_with_instance": True,
        }
        node = self.driver.create_node(
            name=self.name,
            image=self.fake_image,
            size=self.fake_size,
            ex_security_group_id="sg-28ou0f3xa",
            ex_data_disks=self.data_disk,
        )
        self.assertIsNotNone(node)

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()
        self.assertEqual(2, len(sizes))
        size = sizes[0]
        self.assertEqual("ecs.t1.xsmall", size.id)
        self.assertEqual("ecs.t1.xsmall", size.name)
        self.assertEqual(0.5, size.ram)
        self.assertEqual(1, size.extra["cpu_core_count"])
        self.assertEqual("ecs.t1", size.extra["instance_type_family"])
        size = sizes[1]
        self.assertEqual("ecs.s2.small", size.id)
        self.assertEqual("ecs.s2.small", size.name)
        self.assertEqual(1.0, size.ram)
        self.assertEqual(2, size.extra["cpu_core_count"])
        self.assertEqual("ecs.s2", size.extra["instance_type_family"])

    def test_list_locations(self):
        locations = self.driver.list_locations()
        self.assertEqual(9, len(locations))
        location = locations[0]
        self.assertEqual("ap-southeast-1", location.id)
        self.assertIsNone(location.country)

    def test_create_node_without_sg_id_exception(self):
        name = "test_create_node_without_sg_id_exception"
        self.assertRaises(
            AttributeError,
            self.driver.create_node,
            name=name,
            image=self.fake_image,
            size=self.fake_size,
        )

    def test_creat_node_paybytraffic_exception(self):
        name = "test_create_node_paybytraffic_exception"
        self.assertRaises(
            AttributeError,
            self.driver.create_node,
            name=name,
            image=self.fake_image,
            size=self.fake_size,
            ex_security_group_id="sg-id1",
            ex_internet_charge_type="PayByTraffic",
        )

    def test_create_node_ex_system_disk_exception(self):
        name = "test_creat_node_ex_system_disk_exception"
        self.assertRaises(
            AttributeError,
            self.driver.create_node,
            name=name,
            image=self.fake_image,
            size=self.fake_size,
            ex_security_group_id="sg-id1",
            ex_system_disk=None,
        )

    def test_create_node_ex_private_ip_address_exception(self):
        name = "test_create_node_ex_private_ip_address_exception"
        self.assertRaises(
            AttributeError,
            self.driver.create_node,
            name=name,
            image=self.fake_image,
            size=self.fake_size,
            ex_security_group_id="sg-id1",
            ex_private_ip_address="1.1.1.2",
        )

    def test_reboot_node(self):
        ECSMockHttp.type = "reboot_node"
        result = self.driver.reboot_node(self.fake_node)
        self.assertTrue(result)

    def test_reboot_node_with_ex_force_stop(self):
        ECSMockHttp.type = "reboot_node_force_stop"
        result = self.driver.reboot_node(self.fake_node, ex_force_stop=True)
        self.assertTrue(result)

    def test_destroy_node(self):
        ECSMockHttp.type = "destroy_node"
        result = self.driver.destroy_node(self.fake_node)
        self.assertTrue(result)

    def test_ex_start_node(self):
        ECSMockHttp.type = "start_node"
        result = self.driver.ex_start_node(self.fake_node)
        self.assertTrue(result)

    def test_ex_stop_node(self):
        ECSMockHttp.type = "stop_node"
        result = self.driver.ex_stop_node(self.fake_node)
        self.assertTrue(result)

    def test_stop_node_with_ex_force_stop(self):
        ECSMockHttp.type = "stop_node_force_stop"
        result = self.driver.ex_stop_node(self.fake_node, ex_force_stop=True)
        self.assertTrue(result)

    def test_create_public_ip(self):
        ECSMockHttp.type = "create_public_ip"
        result = self.driver.create_public_ip(self.fake_instance_id)
        self.assertTrue(result)

    def test_list_volumes(self):
        volumes = self.driver.list_volumes()
        self.assertEqual(2, len(volumes))
        volume = volumes[0]
        self.assertEqual("d-28m5zbua0", volume.id)
        self.assertEqual("", volume.name)
        self.assertEqual(5, volume.size)
        self.assertEqual(StorageVolumeState.AVAILABLE, volume.state)
        expected_extras = {
            "region_id": "cn-qingdao",
            "zone_id": "cn-qingdao-b",
            "description": "",
            "type": "data",
            "category": "cloud",
            "image_id": "",
            "source_snapshot_id": "",
            "product_code": "",
            "portable": True,
            "instance_id": "",
            "device": "",
            "delete_with_instance": False,
            "enable_auto_snapshot": False,
            "creation_time": "2014-07-23T02:44:07Z",
            "attached_time": "2014-07-23T07:47:35Z",
            "detached_time": "2014-07-23T08:28:48Z",
            "disk_charge_type": "PostPaid",
            "operation_locks": {"lock_reason": None},
        }
        self._validate_extras(expected_extras, volume.extra)
        volume = volumes[1]
        self.assertEqual("d-28zfrmo13", volume.id)
        self.assertEqual("ubuntu1404sys", volume.name)
        self.assertEqual(5, volume.size)
        self.assertEqual(StorageVolumeState.INUSE, volume.state)
        expected_extras = {
            "region_id": "cn-qingdao",
            "zone_id": "cn-qingdao-b",
            "description": "Description",
            "type": "system",
            "category": "cloud",
            "image_id": "ubuntu1404_64_20G_aliaegis_20150325.vhd",
            "source_snapshot_id": "",
            "product_code": "",
            "portable": False,
            "instance_id": "i-28whl2nj2",
            "device": "/dev/xvda",
            "delete_with_instance": True,
            "enable_auto_snapshot": True,
            "creation_time": "2014-07-23T02:44:06Z",
            "attached_time": "2016-01-04T15:02:17Z",
            "detached_time": "",
            "disk_charge_type": "PostPaid",
            "operation_locks": {"lock_reason": None},
        }
        self._validate_extras(expected_extras, volume.extra)

    def test_list_volumes_with_ex_volume_ids(self):
        ECSMockHttp.type = "list_volumes_ex_volume_ids"
        volumes = self.driver.list_volumes(ex_volume_ids=["i-28n7dkvov", "not-existed-id"])
        self.assertIsNotNone(volumes)

    def test_list_volumes_with_ex_filters(self):
        ECSMockHttp.type = "list_volumes_ex_filters"
        ex_filters = {"InstanceId": self.fake_node.id}
        volumes = self.driver.list_volumes(ex_filters=ex_filters)
        self.assertIsNotNone(volumes)

    def test_list_volume_snapshots(self):
        snapshots = self.driver.list_volume_snapshots(self.fake_volume)
        self.assertEqual(1, len(snapshots))

    def test_list_volume_snapshots_with_ex_snapshot_ids(self):
        ECSMockHttp.type = "list_volume_snapshots_ex_snapshot_ids"
        ex_snapshot_ids = ["fake-snapshot1"]
        self.driver.list_volume_snapshots(self.fake_volume, ex_snapshot_ids=ex_snapshot_ids)

    def test_list_volume_snapshots_with_ex_filters(self):
        ECSMockHttp.type = "list_volume_snapshots_ex_filters"
        ex_filters = {"InstanceId": self.fake_node.id}
        self.driver.list_volume_snapshots(self.fake_volume, ex_filters=ex_filters)

    def test_create_volume(self):
        ECSMockHttp.type = "create_volume"
        self.volume_size = 1
        self.volume_name = "fake-volume-name"
        self.description = "fake-description"
        self.disk_category = "system"
        self.client_token = "client_token"
        volume = self.driver.create_volume(
            self.volume_size,
            self.volume_name,
            snapshot=self.fake_snapshot,
            ex_zone_id=self.zone,
            ex_description=self.description,
            ex_disk_category=self.disk_category,
            ex_client_token=self.client_token,
        )
        self.assertIsNotNone(volume)

    def test_create_volume_without_ex_zone_id_exception(self):
        self.assertRaises(AttributeError, self.driver.create_volume, 1, "fake-volume-name")

    def test_create_volume_snapshot(self):
        ECSMockHttp.type = "create_volume_snapshot"
        self.snapshot_name = "fake-snapshot1"
        self.description = "fake-description"
        self.client_token = "client-token"
        snapshot = self.driver.create_volume_snapshot(
            self.fake_volume,
            name=self.snapshot_name,
            ex_description=self.description,
            ex_client_token=self.client_token,
        )
        self.assertIsNotNone(snapshot)

    def test_attach_volume(self):
        self.device = "/dev/sdb"
        self.delete_with_instance = True
        attached = self.driver.attach_volume(
            self.fake_node,
            self.fake_volume,
            device=self.device,
            ex_delete_with_instance=self.delete_with_instance,
        )
        self.assertTrue(attached)

    def test_detach_volume(self):
        self.instance_id = "fake-node1"
        result = self.driver.detach_volume(self.fake_volume, ex_instance_id=self.instance_id)
        self.assertTrue(result)

    def test_detach_volume_query_instance_id(self):
        ECSMockHttp.type = "detach_volume"
        result = self.driver.detach_volume(self.fake_volume)
        self.assertTrue(result)

    def test_detach_volume_query_instance_id_exception(self):
        self.assertRaises(AttributeError, self.driver.detach_volume, self.fake_volume)

    def test_destroy_volume(self):
        ECSMockHttp.type = "destroy_volume"
        result = self.driver.destroy_volume(self.fake_volume)
        self.assertTrue(result)

    def test_destroy_volume_query_volumes_exception(self):
        self.assertRaises(LibcloudError, self.driver.destroy_volume, self.fake_volume)

    def test_destroy_volume_state_exception(self):
        ECSMockHttp.type = "destroy_volume_state"
        self.assertRaises(LibcloudError, self.driver.destroy_volume, self.fake_volume)

    def test_destroy_volume_snapshot(self):
        result = self.driver.destroy_volume_snapshot(self.fake_snapshot)
        self.assertTrue(result)

    def test_destroy_volume_snapshot_exception(self):
        self.assertRaises(AttributeError, self.driver.destroy_volume_snapshot, self.fake_volume)

    def test_list_images(self):
        images = self.driver.list_images(self.fake_location)
        self.assertEqual(1, len(images))
        image = images[0]
        self.assertEqual("freebsd1001_64_20G_aliaegis_20150527.vhd", image.id)
        self.assertEqual("freebsd1001_64_20G_aliaegis_20150527.vhd", image.name)
        expected_extra = {
            "image_version": "1.0.0",
            "os_type": "linux",
            "platform": "Freebsd",
            "architecture": "x86_64",
            "description": "freebsd1001_64_20G_aliaegis_20150527.vhd",
            "size": 20,
            "image_owner_alias": "system",
            "os_name": "FreeBSD  10.1 64位",
            "product_code": "",
            "is_subscribed": False,
            "progress": "100%",
            "creation_time": "2015-06-19T07:25:42Z",
            "usage": "instance",
            "is_copied": False,
        }
        self._validate_extras(expected_extra, image.extra)
        expected_dev_mappings = {
            "snapshot_id": "",
            "size": 20,
            "device": "/dev/xvda",
            "format": "",
            "import_oss_bucket": "",
            "import_oss_object": "",
        }
        self._validate_extras(expected_dev_mappings, image.extra["disk_device_mappings"])

    def test_list_images_with_ex_image_ids(self):
        ECSMockHttp.type = "list_images_ex_image_ids"
        self.driver.list_images(
            location=self.fake_location,
            ex_image_ids=[self.fake_image.id, "not-existed"],
        )

    def test_list_images_with_ex_image_ids_type_exception(self):
        self.assertRaises(
            AttributeError,
            self.driver.list_images,
            location=self.fake_location,
            ex_image_ids={"image_ids": "id1,id2"},
        )

    def test_list_images_with_ex_filters(self):
        ECSMockHttp.type = "list_images_ex_filters"
        ex_filters = {"Status": "Available"}
        self.driver.list_images(location=self.fake_location, ex_filters=ex_filters)

    def test_list_images_multiple_pages(self):
        ECSMockHttp.type = "list_images_pages"
        images = self.driver.list_images()
        self.assertEqual(2, len(images))

    def test_create_image(self):
        self.image_name = "fake-image1"
        self.description = "description"
        self.image_version = "1.0.0"
        self.client_token = "client_token"
        image = self.driver.create_image(
            None,
            self.image_name,
            self.description,
            ex_snapshot_id=self.fake_snapshot.id,
            ex_image_version=self.image_version,
            ex_client_token=self.client_token,
        )
        self.assertIsNotNone(image)

    def test_creaet_image_exception(self):
        self.assertRaises(AttributeError, self.driver.create_image, None, None)

    def test_delete_image(self):
        result = self.driver.delete_image(self.fake_image)
        self.assertTrue(result)

    def test_get_image(self):
        ECSMockHttp.type = "get_image"
        image = self.driver.get_image(self.fake_image.id)
        self.assertIsNotNone(image)

    def test_get_image_not_found_exception(self):
        ECSMockHttp.type = "get_image_not_found"
        self.assertRaises(LibcloudError, self.driver.get_image, self.fake_image.id)

    def test_copy_image(self):
        self.image_name = "copied-image1"
        self.description = "description"
        self.dest_region = "cn-hangzhou"
        self.client_token = "client-token"
        image = self.driver.copy_image(
            self.region,
            self.fake_image,
            self.image_name,
            description=self.description,
            ex_destination_region_id=self.dest_region,
            ex_client_token=self.client_token,
        )
        self.assertIsNotNone(image)

    def test_copy_image_in_the_same_region(self):
        ECSMockHttp.type = "copy_image_same_region"
        image = self.driver.copy_image(self.region, self.fake_image, None)
        self.assertIsNotNone(image)

    def test_ex_create_security_group(self):
        self.sg_description = "description"
        self.client_token = "client-token"
        sg_id = self.driver.ex_create_security_group(
            description=self.sg_description, client_token=self.client_token
        )
        self.assertEqual("sg-F876FF7BA", sg_id)

    def test_ex_list_security_groups(self):
        sgs = self.driver.ex_list_security_groups()
        self.assertEqual(1, len(sgs))
        sg = sgs[0]
        self.assertEqual("sg-28ou0f3xa", sg.id)
        self.assertEqual("sg-28ou0f3xa", sg.name)
        self.assertEqual("System created security group.", sg.description)
        self.assertEqual("", sg.vpc_id)
        self.assertEqual("2015-06-26T08:35:30Z", sg.creation_time)

    def test_ex_join_security_group(self):
        result = self.driver.ex_join_security_group(
            self.fake_node, group_id=self.fake_security_group_id
        )
        self.assertTrue(result)

    def test_ex_leave_security_group(self):
        result = self.driver.ex_leave_security_group(
            self.fake_node, group_id=self.fake_security_group_id
        )
        self.assertTrue(result)

    def test_ex_delete_security_group_by_id(self):
        result = self.driver.ex_delete_security_group_by_id(group_id=self.fake_security_group_id)
        self.assertTrue(result)

    def test_ex_modify_security_group_by_id(self):
        self.sg_name = "name"
        self.sg_description = "description"
        result = self.driver.ex_modify_security_group_by_id(
            group_id=self.fake_security_group_id,
            name=self.sg_name,
            description=self.sg_description,
        )
        self.assertTrue(result)

    def test_ex_list_security_groups_with_ex_filters(self):
        ECSMockHttp.type = "list_sgs_filters"
        self.vpc_id = "vpc1"
        ex_filters = {"VpcId": self.vpc_id}
        sgs = self.driver.ex_list_security_groups(ex_filters=ex_filters)
        self.assertEqual(1, len(sgs))

    def test_ex_list_security_group_attributes(self):
        self.sga_nictype = "internet"
        sgas = self.driver.ex_list_security_group_attributes(
            group_id=self.fake_security_group_id, nic_type=self.sga_nictype
        )
        self.assertEqual(1, len(sgas))
        sga = sgas[0]
        self.assertEqual("ALL", sga.ip_protocol)
        self.assertEqual("-1/-1", sga.port_range)
        self.assertEqual("Accept", sga.policy)
        self.assertEqual("internet", sga.nic_type)

    def test_ex_list_zones(self):
        zones = self.driver.ex_list_zones()
        self.assertEqual(1, len(zones))
        zone = zones[0]
        self.assertEqual("cn-qingdao-b", zone.id)
        self.assertEqual(self.driver, zone.driver)
        self.assertEqual("青岛可用区B", zone.name)
        self.assertIsNotNone(zone.available_resource_types)
        self.assertEqual("IoOptimized", zone.available_resource_types[0])
        self.assertIsNotNone(zone.available_instance_types)
        self.assertEqual("ecs.m2.medium", zone.available_instance_types[0])
        self.assertIsNotNone(zone.available_disk_categories)
        self.assertEqual("cloud_ssd", zone.available_disk_categories[0])


class ECSMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("ecs")

    def _DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _list_nodes_ex_node_ids_DescribeInstances(self, method, url, body, headers):
        params = {"InstanceIds": '["i-28n7dkvov", "not-existed-id"]'}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeInstances(method, url, body, headers)

    def _list_nodes_ex_filters_DescribeInstances(self, method, url, body, headers):
        params = {"ZoneId": self.test.zone}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeInstances(method, url, body, headers)

    def _DescribeInstanceTypes(self, method, url, body, headers):
        resp_body = self.fixtures.load("describe_instance_types.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _DescribeRegions(self, method, url, body, headers):
        resp_body = self.fixtures.load("describe_regions.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_node_CreateInstance(self, method, url, body, headers):
        params = {
            "SecurityGroupId": "sg-28ou0f3xa",
            "Description": "description",
            "InternetChargeType": "PayByTraffic",
            "InternetMaxBandwidthOut": "1",
            "InternetMaxBandwidthIn": "200",
            "HostName": "hostname",
            "Password": "password",
            "IoOptimized": "optimized",
            "SystemDisk.Category": "cloud",
            "SystemDisk.DiskName": "root",
            "SystemDisk.Description": "sys",
            "VSwitchId": "vswitch-id1",
            "PrivateIpAddress": "1.1.1.2",
            "ClientToken": "client_token",
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("create_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_node_DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("create_node_describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_node_StartInstance(self, method, url, body, headers):
        resp_body = self.fixtures.load("start_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_node_with_data_CreateInstance(self, method, url, body, headers):
        params = {
            "SecurityGroupId": "sg-28ou0f3xa",
            "DataDisk.1.Size": "5",
            "DataDisk.1.Category": "cloud",
            "DataDisk.1.DiskName": "data1",
            "DataDisk.1.Description": "description",
            "DataDisk.1.Device": "/dev/xvdb",
            "DataDisk.1.DeleteWithInstance": "true",
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("create_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_node_with_data_DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("create_node_describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_node_with_data_StartInstance(self, method, url, body, headers):
        resp_body = self.fixtures.load("start_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _reboot_node_RebootInstance(self, method, url, body, headers):
        node_id = self.test.fake_node.id
        self.assertUrlContainsQueryParams(url, {"InstanceId": node_id, "ForceStop": "false"})
        resp_body = self.fixtures.load("reboot_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _reboot_node_DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("reboot_node_describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _reboot_node_force_stop_RebootInstance(self, method, url, body, headers):
        node_id = self.test.fake_node.id
        self.assertUrlContainsQueryParams(url, {"InstanceId": node_id, "ForceStop": "true"})
        resp_body = self.fixtures.load("reboot_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _reboot_node_force_stop_DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("reboot_node_describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _destroy_node_DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("destroy_node_describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _destroy_node_DeleteInstance(self, method, url, body, headers):
        node_id = self.test.fake_node.id
        self.assertUrlContainsQueryParams(url, {"InstanceId": node_id})
        resp_body = self.fixtures.load("delete_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _start_node_StartInstance(self, method, url, body, headers):
        node_id = self.test.fake_node.id
        self.assertUrlContainsQueryParams(url, {"InstanceId": node_id})
        resp_body = self.fixtures.load("start_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _start_node_DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("reboot_node_describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _stop_node_StopInstance(self, method, url, body, headers):
        node_id = self.test.fake_node.id
        self.assertUrlContainsQueryParams(url, {"InstanceId": node_id, "ForceStop": "false"})
        resp_body = self.fixtures.load("stop_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _stop_node_DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("stop_node_describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _stop_node_force_stop_StopInstance(self, method, url, body, headers):
        node_id = self.test.fake_node.id
        self.assertUrlContainsQueryParams(url, {"InstanceId": node_id, "ForceStop": "true"})
        resp_body = self.fixtures.load("stop_instance.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _stop_node_force_stop_DescribeInstances(self, method, url, body, headers):
        resp_body = self.fixtures.load("stop_node_describe_instances.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _DescribeDisks(self, method, url, body, headers):
        resp_body = self.fixtures.load("describe_disks.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _list_volumes_ex_volume_ids_DescribeDisks(self, method, url, body, headers):
        region = self.test.region
        params = {"DiskIds": '["i-28n7dkvov", "not-existed-id"]', "RegionId": region}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeInstances(method, url, body, headers)

    def _list_volumes_ex_filters_DescribeDisks(self, method, url, body, headers):
        params = {"InstanceId": self.test.fake_node.id}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeDisks(method, url, body, headers)

    def _DescribeSnapshots(self, method, url, body, headers):
        region = self.test.region
        volume_id = self.test.fake_volume.id
        params = {"RegionId": region, "DiskId": volume_id}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("describe_snapshots.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _list_volume_snapshots_ex_snapshot_ids_DescribeSnapshots(self, method, url, body, headers):
        params = {"RegionId": self.test.region, "SnapshotIds": '["fake-snapshot1"]'}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeSnapshots(method, url, body, headers)

    def _list_volume_snapshots_ex_filters_DescribeSnapshots(self, method, url, body, headers):
        params = {"InstanceId": self.test.fake_node.id}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeSnapshots(method, url, body, headers)

    def _create_volume_CreateDisk(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "DiskName": self.test.volume_name,
            "Size": str(self.test.volume_size),
            "ZoneId": self.test.zone,
            "SnapshotId": self.test.fake_snapshot.id,
            "Description": self.test.description,
            "DiskCategory": self.test.disk_category,
            "ClientToken": self.test.client_token,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("create_disk.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_volume_DescribeDisks(self, method, url, body, headers):
        resp_body = self.fixtures.load("create_volume_describe_disks.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_volume_snapshot_CreateSnapshot(self, method, url, body, headers):
        params = {
            "DiskId": self.test.fake_volume.id,
            "SnapshotName": self.test.snapshot_name,
            "Description": self.test.description,
            "ClientToken": self.test.client_token,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("create_snapshot.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_volume_snapshot_DescribeSnapshots(self, method, url, body, headers):
        resp_body = self.fixtures.load("describe_snapshots.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _AttachDisk(self, method, url, body, headers):
        delete_with_instance = str(self.test.delete_with_instance).lower()
        params = {
            "InstanceId": self.test.fake_node.id,
            "DiskId": self.test.fake_volume.id,
            "Device": self.test.device,
            "DeleteWithInstance": delete_with_instance,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("attach_disk.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _DetachDisk(self, method, url, body, headers):
        params = {
            "DiskId": self.test.fake_volume.id,
            "InstanceId": self.test.instance_id,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("detach_disk.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _detach_volume_DescribeDisks(self, method, url, body, headers):
        params = {"DiskIds": '["' + self.test.fake_volume.id + '"]'}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("detach_volume_describe_disks.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _detach_volume_DetachDisk(self, method, url, body, headers):
        params = {"DiskId": self.test.fake_volume.id, "InstanceId": "i-28whl2nj2"}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("detach_disk.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _destroy_volume_DescribeDisks(self, method, url, body, headers):
        params = {"DiskIds": '["' + self.test.fake_volume.id + '"]'}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("destroy_volume_describe_disks.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _destroy_volume_DeleteDisk(self, method, url, body, headers):
        params = {"DiskId": self.test.fake_volume.id}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("delete_disk.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _destroy_volume_state_DescribeDisks(self, method, url, body, headers):
        return self._detach_volume_DescribeDisks(method, url, body, headers)

    def _DeleteSnapshot(self, method, url, body, header):
        params = {"SnapshotId": self.test.fake_snapshot.id}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("delete_snapshot.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _DescribeImages(self, method, url, body, headers):
        params = {"RegionId": self.test.fake_location.id}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("describe_images.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _list_images_pages_DescribeImages(self, method, url, body, headers):
        if "PageNumber=2" in url:
            resp_body = self.fixtures.load("pages_describe_images_page2.xml")
        else:
            resp_body = self.fixtures.load("pages_describe_images.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _list_images_ex_image_ids_DescribeImages(self, method, url, body, headers):
        params = {"ImageId": self.test.fake_image.id + ",not-existed"}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeImages(method, url, body, headers)

    def _list_images_ex_filters_DescribeImages(self, method, url, body, headers):
        params = {"Status": "Available"}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeImages(method, url, body, headers)

    def _CreateImage(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "ImageName": self.test.image_name,
            "Description": self.test.description,
            "SnapshotId": self.test.fake_snapshot.id,
            "ImageVersion": self.test.image_version,
            "ClientToken": self.test.client_token,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("create_image.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _DeleteImage(self, method, url, body, headers):
        params = {"RegionId": self.test.region, "ImageId": self.test.fake_image.id}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("delete_image.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _get_image_DescribeImages(self, method, url, body, headers):
        params = {"RegionId": self.test.region, "ImageId": self.test.fake_image.id}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("describe_images.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _get_image_not_found_DescribeImages(self, method, url, body, headers):
        params = {"RegionId": self.test.region, "ImageId": self.test.fake_image.id}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("get_image_describe_images.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _CopyImage(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "ImageId": self.test.fake_image.id,
            "DestinationRegionId": self.test.dest_region,
            "DestinationImageName": self.test.image_name,
            "DestinationDescription": self.test.description,
            "ClientToken": self.test.client_token,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("copy_image.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _copy_image_same_region_CopyImage(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "ImageId": self.test.fake_image.id,
            "DestinationRegionId": self.test.region,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("copy_image.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _copy_image_same_region_DescribeImages(self, method, url, body, headers):
        return self._DescribeImages(method, url, body, headers)

    def _DescribeSecurityGroups(self, method, url, body, headers):
        params = {"RegionId": self.test.region}
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("describe_security_groups.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _JoinSecurityGroup(self, method, url, body, headers):
        params = {
            "InstanceId": self.test.fake_node.id,
            "SecurityGroupId": self.test.fake_security_group_id,
        }
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("join_security_group_by_id.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _LeaveSecurityGroup(self, method, url, body, headers):
        params = {
            "InstanceId": self.test.fake_node.id,
            "SecurityGroupId": self.test.fake_security_group_id,
        }
        self.assertUrlContainsQueryParams(url, params)
        body = self.fixtures.load("leave_security_group_by_id.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _list_sgs_filters_DescribeSecurityGroups(self, method, url, body, headers):
        params = {"VpcId": self.test.vpc_id}
        self.assertUrlContainsQueryParams(url, params)
        return self._DescribeSecurityGroups(method, url, body, headers)

    def _CreateSecurityGroup(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "Description": self.test.sg_description,
            "ClientToken": self.test.client_token,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("create_security_group.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _DeleteSecurityGroup(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "SecurityGroupId": self.test.fake_security_group_id,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("delete_security_group_by_id.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _ModifySecurityGroupAttribute(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "SecurityGroupId": self.test.fake_security_group_id,
            "SecurityGroupName": self.test.sg_name,
            "Description": self.test.sg_description,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("modify_security_group_by_id.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _DescribeSecurityGroupAttribute(self, method, url, body, headers):
        params = {
            "RegionId": self.test.region,
            "SecurityGroupId": self.test.fake_security_group_id,
            "NicType": self.test.sga_nictype,
        }
        self.assertUrlContainsQueryParams(url, params)
        resp_body = self.fixtures.load("describe_security_group_attributes.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _DescribeZones(self, method, url, body, headers):
        resp_body = self.fixtures.load("describe_zones.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])

    def _create_public_ip_AllocatePublicIpAddress(self, method, url, body, headers):
        resp_body = self.fixtures.load("create_public_ip.xml")
        return (httplib.OK, resp_body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common utilities for OpenStack
"""

from libcloud.utils.py3 import ET, httplib
from libcloud.common.base import Response, ConnectionUserAndKey
from libcloud.common.types import ProviderError
from libcloud.compute.types import LibcloudError, MalformedResponseError, KeyPairDoesNotExistError
from libcloud.common.exceptions import BaseHTTPError

# Imports for backward compatibility reasons
from libcloud.common.openstack_identity import (
    AUTH_TOKEN_HEADER,
    OpenStackServiceCatalog,
    OpenStackIdentityTokenScope,
    get_class_for_auth_version,
)

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore

AUTH_API_VERSION = "1.1"

# Auth versions which contain token expiration information.
AUTH_VERSIONS_WITH_EXPIRES = [
    "1.1",
    "2.0",
    "2.0_apikey",
    "2.0_password",
    "3.x",
    "3.x_password",
]

__all__ = [
    "OpenStackBaseConnection",
    "OpenStackResponse",
    "OpenStackException",
    "OpenStackDriverMixin",
]


class OpenStackBaseConnection(ConnectionUserAndKey):

    """
    Base class for OpenStack connections.

    :param user_id: User name to use when authenticating
    :type user_id: ``str``

    :param key: Secret to use when authenticating.
    :type key: ``str``

    :param secure: Use HTTPS?  (True by default.)
    :type secure: ``bool``

    :param ex_force_base_url: Base URL for connection requests.  If
                              not specified, this will be determined by
                              authenticating.
    :type ex_force_base_url: ``str``

    :param ex_force_auth_url: Base URL for authentication requests.
    :type ex_force_auth_url: ``str``

    :param ex_force_auth_version: Authentication version to use.  If
                                  not specified, defaults to AUTH_API_VERSION.
    :type ex_force_auth_version: ``str``

    :param ex_force_auth_token: Authentication token to use for connection
                                requests.  If specified, the connection will
                                not attempt to authenticate, and the value
                                of ex_force_base_url will be used to
                                determine the base request URL.  If
                                ex_force_auth_token is passed in,
                                ex_force_base_url must also be provided.
    :type ex_force_auth_token: ``str``

    :param token_scope: Whether to scope a token to a "project", a
                        "domain" or "unscoped".
    :type token_scope: ``str``

    :param ex_domain_name: When authenticating, provide this domain name to
                           the identity service.  A scoped token will be
                           returned. Some cloud providers require the domain
                           name to be provided at authentication time. Others
                           will use a default domain if none is provided.
    :type ex_domain_name: ``str``

    :param ex_tenant_name: When authenticating, provide this tenant name to the
                           identity service. A scoped token will be returned.
                           Some cloud providers require the tenant name to be
                           provided at authentication time. Others will use a
                           default tenant if none is provided.
    :type ex_tenant_name: ``str``

    :param ex_tenant_domain_id: When authenticating, provide this tenant
                                domain id to the identity service.
                                A scoped token will be returned.
                                Some cloud providers require the tenant
                                domain id to be provided at authentication
                                time. Others will use a default tenant
                                domain id if none is provided.
    :type ex_tenant_domain_id: ``str``

    :param ex_force_service_type: Service type to use when selecting an
                                  service. If not specified, a provider
                                  specific default will be used.
    :type ex_force_service_type: ``str``

    :param ex_force_service_name: Service name to use when selecting an
                                  service. If not specified, a provider
                                  specific default will be used.
    :type ex_force_service_name: ``str``

    :param ex_force_service_region: Region to use when selecting an service.
                                    If not specified, a provider specific
                                    default will be used.
    :type ex_force_service_region: ``str``

    :param ex_auth_cache: External cache where authentication tokens are
                          stored for reuse by other processes. Tokens are
                          always cached in memory on the driver instance. To
                          share tokens among multiple drivers, processes, or
                          systems, pass a cache here.
    :type ex_auth_cache: :class:`OpenStackAuthenticationCache`
    """

    auth_url = None  # type: str
    auth_token = None  # type: str
    auth_token_expires = None
    auth_user_info = None
    service_catalog = None
    service_type = None
    service_name = None
    service_region = None
    accept_format = None
    _auth_version = None  # type: str

    def __init__(
        self,
        user_id,
        key,
        secure=True,
        host=None,
        port=None,
        timeout=None,
        proxy_url=None,
        ex_force_base_url=None,
        ex_force_auth_url=None,
        ex_force_auth_version=None,
        ex_force_auth_token=None,
        ex_token_scope=OpenStackIdentityTokenScope.PROJECT,
        ex_domain_name="Default",
        ex_tenant_name=None,
        ex_tenant_domain_id="default",
        ex_force_service_type=None,
        ex_force_service_name=None,
        ex_force_service_region=None,
        ex_force_microversion=None,
        ex_auth_cache=None,
        retry_delay=None,
        backoff=None,
    ):
        super().__init__(
            user_id,
            key,
            secure=secure,
            timeout=timeout,
            retry_delay=retry_delay,
            backoff=backoff,
            proxy_url=proxy_url,
        )

        if ex_force_auth_version:
            self._auth_version = ex_force_auth_version

        self.base_url = ex_force_base_url
        self._ex_force_base_url = ex_force_base_url
        self._ex_force_auth_url = ex_force_auth_url
        self._ex_force_auth_token = ex_force_auth_token
        self._ex_token_scope = ex_token_scope
        self._ex_domain_name = ex_domain_name
        self._ex_tenant_name = ex_tenant_name
        self._ex_tenant_domain_id = ex_tenant_domain_id
        self._ex_force_service_type = ex_force_service_type
        self._ex_force_service_name = ex_force_service_name
        self._ex_force_service_region = ex_force_service_region
        self._ex_force_microversion = ex_force_microversion
        self._ex_auth_cache = ex_auth_cache
        self._osa = None

        if ex_force_auth_token and not ex_force_base_url:
            raise LibcloudError(
                "Must also provide ex_force_base_url when specifying " "ex_force_auth_token."
            )

        if ex_force_auth_token:
            self.auth_token = ex_force_auth_token

        if not self._auth_version:
            self._auth_version = AUTH_API_VERSION

        auth_url = self._get_auth_url()

        if not auth_url:
            raise LibcloudError("OpenStack instance must " + "have auth_url set")

    def get_auth_class(self):
        """
        Retrieve identity / authentication class instance.

        :rtype: :class:`OpenStackIdentityConnection`
        """
        if not self._osa:
            auth_url = self._get_auth_url()

            cls = get_class_for_auth_version(auth_version=self._auth_version)
            self._osa = cls(
                auth_url=auth_url,
                user_id=self.user_id,
                key=self.key,
                tenant_name=self._ex_tenant_name,
                tenant_domain_id=self._ex_tenant_domain_id,
                domain_name=self._ex_domain_name,
                token_scope=self._ex_token_scope,
                timeout=self.timeout,
                proxy_url=self.proxy_url,
                parent_conn=self,
                auth_cache=self._ex_auth_cache,
            )

        return self._osa

    def request(self, action, params=None, data="", headers=None, method="GET", raw=False):
        headers = headers or {}
        params = params or {}

        # Include default content-type for POST and PUT request (if available)
        default_content_type = getattr(self, "default_content_type", None)
        if method.upper() in ["POST", "PUT"] and default_content_type:
            headers["Content-Type"] = default_content_type

        try:
            return super().request(
                action=action,
                params=params,
                data=data,
                method=method,
                headers=headers,
                raw=raw,
            )
        except BaseHTTPError as ex:
            # Evict cached auth token if we receive Unauthorized while using it
            if ex.code == httplib.UNAUTHORIZED and self._ex_force_auth_token is None:
                self.get_auth_class().clear_cached_auth_context()
            raise

    def _get_auth_url(self):
        """
        Retrieve auth url for this instance using either "ex_force_auth_url"
        constructor kwarg of "auth_url" class variable.
        """
        auth_url = self.auth_url

        if self._ex_force_auth_url is not None:
            auth_url = self._ex_force_auth_url

        return auth_url

    def get_service_catalog(self):
        if self.service_catalog is None:
            self._populate_hosts_and_request_paths()

        return self.service_catalog

    def get_service_name(self):
        """
        Gets the service name used to look up the endpoint in the service
        catalog.

        :return: name of the service in the catalog
        """
        if self._ex_force_service_name:
            return self._ex_force_service_name

        return self.service_name

    def get_endpoint(self):
        """
        Selects the endpoint to use based on provider specific values,
        or overrides passed in by the user when setting up the driver.

        :returns: url of the relevant endpoint for the driver
        """
        service_type = self.service_type
        service_name = self.service_name
        service_region = self.service_region

        if self._ex_force_service_type:
            service_type = self._ex_force_service_type
        if self._ex_force_service_name:
            service_name = self._ex_force_service_name
        if self._ex_force_service_region:
            service_region = self._ex_force_service_region

        endpoint = self.service_catalog.get_endpoint(
            service_type=service_type, name=service_name, region=service_region
        )

        url = endpoint.url

        if not url:
            raise LibcloudError("Could not find specified endpoint")

        return url

    def add_default_headers(self, headers):
        headers[AUTH_TOKEN_HEADER] = self.auth_token
        headers["Accept"] = self.accept_format
        if self._ex_force_microversion:
            # If service not set in microversion, assume compute
            microversion = self._ex_force_microversion.strip().split()
            if len(microversion) == 2:
                service_type = microversion[0]
                microversion = microversion[1]
            elif len(microversion) == 1:
                service_type = "compute"
                microversion = microversion[0]
            else:
                raise LibcloudError("Invalid microversion format: servicename X.XX")

            if self.service_type and self.service_type.startswith(service_type):
                headers["OpenStack-API-Version"] = "{} {}".format(
                    service_type,
                    microversion,
                )
        return headers

    def morph_action_hook(self, action):
        self._populate_hosts_and_request_paths()
        return super().morph_action_hook(action)

    def _set_up_connection_info(self, url):
        prev_conn = (self.host, self.port, self.secure)
        result = self._tuple_from_url(url)
        (self.host, self.port, self.secure, self.request_path) = result
        new_conn = (self.host, self.port, self.secure)
        if new_conn != prev_conn:
            # We only call connect in case connection details have changed - this way we correctly
            # re-use connection in case nothing has changed
            self.connect()

    def _populate_hosts_and_request_paths(self):
        """
        OpenStack uses a separate host for API calls which is only provided
        after an initial authentication request.
        """
        osa = self.get_auth_class()

        if self._ex_force_auth_token:
            # If ex_force_auth_token is provided we always hit the api directly
            # and never try to authenticate.
            #
            # Note: When ex_force_auth_token is provided, ex_force_base_url
            # must be provided as well.
            self._set_up_connection_info(url=self._ex_force_base_url)
            return

        if not osa.is_token_valid():
            # Token is not available or it has expired. Need to retrieve a
            # new one.
            if self._auth_version == "2.0_apikey":
                kwargs = {"auth_type": "api_key"}
            elif self._auth_version == "2.0_password":
                kwargs = {"auth_type": "password"}
            else:
                kwargs = {}

            # pylint: disable=unexpected-keyword-arg
            osa = osa.authenticate(**kwargs)  # may throw InvalidCreds
            # pylint: enable=unexpected-keyword-arg

            self.auth_token = osa.auth_token
            self.auth_token_expires = osa.auth_token_expires
            self.auth_user_info = osa.auth_user_info

            # Pull out and parse the service catalog
            osc = OpenStackServiceCatalog(service_catalog=osa.urls, auth_version=self._auth_version)
            self.service_catalog = osc

        url = self._ex_force_base_url or self.get_endpoint()
        self._set_up_connection_info(url=url)


class OpenStackException(ProviderError):
    pass


class OpenStackResponse(Response):
    node_driver = None

    def success(self):
        i = int(self.status)
        return 200 <= i <= 299

    def has_content_type(self, content_type):
        content_type_value = self.headers.get("content-type") or ""
        content_type_value = content_type_value.lower()
        return content_type_value.find(content_type.lower()) > -1

    def parse_body(self):
        if self.status == httplib.NO_CONTENT or not self.body:
            return None

        if self.has_content_type("application/xml"):
            try:
                return ET.XML(self.body)
            except Exception:
                raise MalformedResponseError(
                    "Failed to parse XML", body=self.body, driver=self.node_driver
                )

        elif self.has_content_type("application/json"):
            try:
                return json.loads(self.body)
            except Exception:
                raise MalformedResponseError(
                    "Failed to parse JSON", body=self.body, driver=self.node_driver
                )
        else:
            return self.body

    def parse_error(self):
        body = self.parse_body()

        if self.has_content_type("application/xml"):
            text = "; ".join([err.text or "" for err in body.getiterator() if err.text])
        elif self.has_content_type("application/json"):
            values = list(body.values())

            context = self.connection.context
            driver = self.connection.driver
            key_pair_name = context.get("key_pair_name", None)

            if (
                len(values) > 0
                and "code" in values[0]
                and values[0]["code"] == 404
                and key_pair_name
            ):
                raise KeyPairDoesNotExistError(name=key_pair_name, driver=driver)
            elif len(values) > 0 and "message" in values[0]:
                text = ";".join([fault_data["message"] for fault_data in values])
            else:
                text = body
        else:
            # while we hope a response is always one of xml or json, we have
            # seen html or text in the past, its not clear we can really do
            # something to make it more readable here, so we will just pass
            # it along as the whole response body in the text variable.
            text = body

        return "{} {} {}".format(self.status, self.error, text)


class OpenStackDriverMixin:
    def __init__(
        self,
        ex_force_base_url=None,
        ex_force_auth_url=None,
        ex_force_auth_version=None,
        ex_force_auth_token=None,
        ex_token_scope=OpenStackIdentityTokenScope.PROJECT,
        ex_domain_name="Default",
        ex_tenant_name=None,
        ex_tenant_domain_id="default",
        ex_force_service_type=None,
        ex_force_service_name=None,
        ex_force_service_region=None,
        ex_auth_cache=None,
        ex_force_microversion=None,
        *args,
        **kwargs,
    ):
        self._ex_force_base_url = ex_force_base_url
        self._ex_force_auth_url = ex_force_auth_url
        self._ex_force_auth_version = ex_force_auth_version
        self._ex_force_auth_token = ex_force_auth_token
        self._ex_token_scope = ex_token_scope
        self._ex_domain_name = ex_domain_name
        self._ex_tenant_name = ex_tenant_name
        self._ex_tenant_domain_id = ex_tenant_domain_id
        self._ex_force_service_type = ex_force_service_type
        self._ex_force_service_name = ex_force_service_name
        self._ex_force_service_region = ex_force_service_region
        self._ex_auth_cache = ex_auth_cache
        self._ex_force_microversion = ex_force_microversion

    def openstack_connection_kwargs(self):
        """
        Returns certain ``ex_*`` parameters for this connection.

        :rtype: ``dict``
        """
        rv = {}
        if self._ex_force_base_url:
            rv["ex_force_base_url"] = self._ex_force_base_url
        if self._ex_force_auth_token:
            rv["ex_force_auth_token"] = self._ex_force_auth_token
        if self._ex_force_auth_url:
            rv["ex_force_auth_url"] = self._ex_force_auth_url
        if self._ex_force_auth_version:
            rv["ex_force_auth_version"] = self._ex_force_auth_version
        if self._ex_token_scope:
            rv["ex_token_scope"] = self._ex_token_scope
        if self._ex_domain_name:
            rv["ex_domain_name"] = self._ex_domain_name
        if self._ex_tenant_name:
            rv["ex_tenant_name"] = self._ex_tenant_name
        if self._ex_tenant_domain_id:
            rv["ex_tenant_domain_id"] = self._ex_tenant_domain_id
        if self._ex_force_service_type:
            rv["ex_force_service_type"] = self._ex_force_service_type
        if self._ex_force_service_name:
            rv["ex_force_service_name"] = self._ex_force_service_name
        if self._ex_force_service_region:
            rv["ex_force_service_region"] = self._ex_force_service_region
        if self._ex_auth_cache is not None:
            rv["ex_auth_cache"] = self._ex_auth_cache
        if self._ex_force_microversion:
            rv["ex_force_microversion"] = self._ex_force_microversion
        return rv



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import json
import time
import base64
import unittest
from unittest import mock

import libcloud.common.gig_g8
from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.compute.base import Node, NodeSize, NodeImage, StorageVolume
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.gig_g8 import G8Network, G8NodeDriver, G8PortForward

# For tests we don't want to actually verify a token since we use an expired on
original_is_jwt_expired = libcloud.common.gig_g8.is_jwt_expired


def mock_is_jwt_expired(jwt):
    return False


class G8MockHttp(MockHttp):
    """Fixtures needed for tests related to rating model"""

    fixtures = ComputeFileFixtures("gig_g8")

    def __getattr__(self, key):
        def method(method, path, params, headers):
            response = self.fixtures.load("{}_{}.json".format(method, key.lstrip("_")))
            return (httplib.OK, response, {}, httplib.responses[httplib.OK])

        return method


@mock.patch("libcloud.common.gig_g8.is_jwt_expired", mock_is_jwt_expired)
class G8Tests(unittest.TestCase):
    def setUp(self):
        G8NodeDriver.connectionCls.conn_class = G8MockHttp
        self.driver = G8NodeDriver(
            1,
            "header.eyJhenAiOiJkZndlcmdyZWdyZSIsImV4cCI6MTU5MDUyMzEwNSwiaXNzIjoiaXRzeW91b25saW5lIiwicmVmcmVzaF90b2tlbiI6Inh4eHh4eHgiLCJzY29wZSI6WyJ1c2VyOmFkbWluIl0sInVzZXJuYW1lIjoiZXhhbXBsZSJ9.signature",
            "https://myg8.example.com",
        )

    def test_list_networks(self):
        networks = self.driver.ex_list_networks()
        self.assertGreater(len(networks), 0)
        for network in networks:
            self.assertIsInstance(network, G8Network)

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        self.assertGreater(len(nodes), 0)
        for node in nodes:
            self.assertIsInstance(node, Node)

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()
        self.assertGreater(len(sizes), 0)
        for size in sizes:
            self.assertIsInstance(size, NodeSize)

    def test_list_images(self):
        images = self.driver.list_images()
        self.assertGreater(len(images), 0)
        for image in images:
            self.assertIsInstance(image, NodeImage)

    def test_create_node(self):
        image = self.driver.list_images()[0]
        size = self.driver.list_sizes()[0]
        network = self.driver.ex_list_networks()[0]
        node = self.driver.create_node("my test", image, network, "my description", size)
        self.assertIsInstance(node, Node)

    def test_stop_node(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.stop_node(node)
        self.assertTrue(result)

    def test_reboot_node(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.reboot_node(node)
        self.assertTrue(result)

    def test_destroy_node(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.destroy_node(node)
        self.assertTrue(result)

    def test_start_node(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.start_node(node)
        self.assertTrue(result)

    def test_create_network(self):
        network = self.driver.ex_create_network("my network")
        self.assertIsInstance(network, G8Network)

    def test_destroy_network(self):
        network = self.driver.ex_list_networks()[0]
        result = self.driver.ex_destroy_network(network)
        self.assertTrue(result)

    def test_list_portforward(self):
        network = self.driver.ex_list_networks()[0]
        forwards = self.driver.ex_list_portforwards(network)
        self.assertGreater(len(forwards), 0)
        for forward in forwards:
            self.assertIsInstance(forward, G8PortForward)

    def test_create_forward(self):
        network = self.driver.ex_list_networks()[0]
        node = self.driver.list_nodes()[0]
        result = self.driver.ex_create_portforward(network, node, 34022, 22)
        self.assertIsInstance(result, G8PortForward)

    def test_delete_portforward(self):
        network = self.driver.ex_list_networks()[0]
        forward = self.driver.ex_list_portforwards(network)[0]
        res = self.driver.ex_delete_portforward(forward)
        self.assertTrue(res)

    def test_list_volumes(self):
        volumes = self.driver.list_volumes()
        self.assertGreater(len(volumes), 1)
        for volume in volumes:
            self.assertIsInstance(volume, StorageVolume)

    def test_create_volume(self):
        volume = self.driver.create_volume(10, "my volume", "my description")
        self.assertIsInstance(volume, StorageVolume)

    def test_destroy_volume(self):
        volume = self.driver.list_volumes()[0]
        res = self.driver.destroy_volume(volume)
        self.assertTrue(res)

    def test_attach_volume(self):
        volume = self.driver.list_volumes()[0]
        node = self.driver.list_nodes()[0]
        res = self.driver.attach_volume(node, volume)
        self.assertTrue(res)

    def test_detach_volume(self):
        volume = self.driver.list_volumes()[0]
        node = self.driver.list_nodes()[0]
        res = self.driver.detach_volume(node, volume)
        self.assertTrue(res)

    def test_is_jwt_expired(self):
        data = {
            "azp": "example",
            "exp": int(time.time()),
            "iss": "itsyouonline",
            "refresh_token": "xxxxxxx",
            "scope": ["user:admin"],
            "username": "example",
        }

        def contruct_jwt(data):
            jsondata = json.dumps(data).encode()
            return "header.{}.signature".format(base64.encodebytes(jsondata).decode())

        libcloud.common.gig_g8.is_jwt_expired = original_is_jwt_expired

        self.assertTrue(libcloud.common.gig_g8.is_jwt_expired(contruct_jwt(data)))
        data["exp"] = int(time.time()) + 300  # expire in 5min
        self.assertFalse(libcloud.common.gig_g8.is_jwt_expired(contruct_jwt(data)))


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Maintainer: Jed Smith <jed@linode.com>
# Based upon code written by Alex Polvi <polvi@cloudkick.com>
#

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.compute.base import Node, StorageVolume, NodeAuthSSHKey, NodeAuthPassword
from libcloud.test.compute import TestCaseMixin
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.linode import LinodeNodeDriver


class LinodeTest(unittest.TestCase, TestCaseMixin):
    # The Linode test suite

    def setUp(self):
        LinodeNodeDriver.connectionCls.conn_class = LinodeMockHttp
        LinodeMockHttp.use_param = "api_action"
        self.driver = LinodeNodeDriver("foo", api_version="3.0")

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(len(nodes), 1)
        node = nodes[0]
        self.assertEqual(node.id, "8098")
        self.assertEqual(node.name, "api-node3")
        self.assertEqual(node.extra["PLANID"], "2")
        self.assertTrue("75.127.96.245" in node.public_ips)
        self.assertEqual(node.private_ips, [])

    def test_reboot_node(self):
        # An exception would indicate failure
        node = self.driver.list_nodes()[0]
        self.driver.reboot_node(node)

    def test_destroy_node(self):
        # An exception would indicate failure
        node = self.driver.list_nodes()[0]
        self.driver.destroy_node(node)

    def test_create_node_password_auth(self):
        # Will exception on failure
        self.driver.create_node(
            name="Test",
            location=self.driver.list_locations()[0],
            size=self.driver.list_sizes()[0],
            image=self.driver.list_images()[6],
            auth=NodeAuthPassword("test123"),
        )

    def test_create_node_ssh_key_auth(self):
        # Will exception on failure
        node = self.driver.create_node(
            name="Test",
            location=self.driver.list_locations()[0],
            size=self.driver.list_sizes()[0],
            image=self.driver.list_images()[6],
            auth=NodeAuthSSHKey("foo"),
        )
        self.assertTrue(isinstance(node, Node))

    def test_list_sizes(self):
        sizes = self.driver.list_sizes()
        self.assertEqual(len(sizes), 9)
        for size in sizes:
            self.assertEqual(size.ram, int(size.name.split(" ")[1]))

    def test_list_images(self):
        images = self.driver.list_images()
        self.assertEqual(len(images), 30)

    def test_create_node_response(self):
        # should return a node object
        node = self.driver.create_node(
            name="node-name",
            location=self.driver.list_locations()[0],
            size=self.driver.list_sizes()[0],
            image=self.driver.list_images()[0],
            auth=NodeAuthPassword("foobar"),
        )
        self.assertTrue(isinstance(node, Node))

    def test_destroy_volume(self):
        # Will exception on failure
        node = self.driver.list_nodes()[0]
        volume = StorageVolume(
            id=55648,
            name="test",
            size=1024,
            driver=self.driver,
            extra={"LINODEID": node.id},
        )
        self.driver.destroy_volume(volume)

    def test_ex_create_volume(self):
        # should return a StorageVolume object
        node = self.driver.list_nodes()[0]
        volume = self.driver.ex_create_volume(
            size=4096, name="Another test image", node=node, fs_type="ext4"
        )
        self.assertTrue(isinstance(volume, StorageVolume))

    def test_ex_list_volumes(self):
        # should return list of StorageVolume objects
        node = self.driver.list_nodes()[0]
        volumes = self.driver.ex_list_volumes(node=node)

        self.assertTrue(isinstance(volumes, list))
        self.assertTrue(isinstance(volumes[0], StorageVolume))
        self.assertEqual(len(volumes), 2)


class LinodeMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("linode")

    def _avail_datacenters(self, method, url, body, headers):
        body = self.fixtures.load("_avail_datacenters.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _avail_linodeplans(self, method, url, body, headers):
        body = self.fixtures.load("_avail_linodeplans.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _avail_distributions(self, method, url, body, headers):
        body = self.fixtures.load("_avail_distributions.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_create(self, method, url, body, headers):
        body = '{"ERRORARRAY":[],"ACTION":"linode.create","DATA":{"LinodeID":8098}}'
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_disk_create(self, method, url, body, headers):
        body = (
            '{"ERRORARRAY":[],"ACTION":"linode.disk.create","DATA":{"JobID":1298,"DiskID":55647}}'
        )
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_disk_delete(self, method, url, body, headers):
        body = (
            '{"ERRORARRAY":[],"ACTION":"linode.disk.delete","DATA":{"JobID":1298,"DiskID":55648}}'
        )
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_disk_createfromdistribution(self, method, url, body, headers):
        body = '{"ERRORARRAY":[],"ACTION":"linode.disk.createFromDistribution","DATA":{"JobID":1298,"DiskID":55647}}'
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_disk_list(self, method, url, body, headers):
        body = self.fixtures.load("_linode_disk_list.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_delete(self, method, url, body, headers):
        body = '{"ERRORARRAY":[],"ACTION":"linode.delete","DATA":{"LinodeID":8098}}'
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_update(self, method, url, body, headers):
        body = '{"ERRORARRAY":[],"ACTION":"linode.update","DATA":{"LinodeID":8098}}'
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_reboot(self, method, url, body, headers):
        body = '{"ERRORARRAY":[],"ACTION":"linode.reboot","DATA":{"JobID":1305}}'
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _avail_kernels(self, method, url, body, headers):
        body = self.fixtures.load("_avail_kernels.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_boot(self, method, url, body, headers):
        body = '{"ERRORARRAY":[],"ACTION":"linode.boot","DATA":{"JobID":1300}}'
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_config_create(self, method, url, body, headers):
        body = '{"ERRORARRAY":[],"ACTION":"linode.config.create","DATA":{"ConfigID":31239}}'
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_list(self, method, url, body, headers):
        body = self.fixtures.load("_linode_list.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _linode_ip_list(self, method, url, body, headers):
        body = self.fixtures.load("_linode_ip_list.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _batch(self, method, url, body, headers):
        body = self.fixtures.load("_batch.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from datetime import datetime

from libcloud.test import LibcloudTestCase
from libcloud.common.osc import OSCRequestSignerAlgorithmV4
from libcloud.compute.providers import Provider, get_driver


class OSCRequestSignerAlgorithmV4TestCase(LibcloudTestCase):
    def setUp(self):
        cls = get_driver(Provider.OUTSCALE)
        self.driver = cls(
            key="my_key", secret="my_secret", region="my_region", service="my_service"
        )
        self.now = datetime(2015, 3, 4, hour=17, minute=34, second=52)
        self.version = "latest"
        self.signer = OSCRequestSignerAlgorithmV4(
            access_key=self.driver.key,
            access_secret=self.driver.secret,
            version=self.version,
            connection=self.driver.connection,
        )

    def test_v4_signature_contains_user_id(self):
        action = "ReadImages"
        headers = self.signer.get_request_headers(
            action=action,
            data="{}",
            service_name=self.driver.service_name,
            region=self.driver.region,
        )
        self.assertIn("Credential=my_key/", headers["Authorization"])

    def test_v4_signature_contains_credential_scope(self):
        action = "ReadImages"
        headers = self.signer.get_request_headers(
            action=action,
            data="{}",
            service_name=self.driver.service_name,
            region=self.driver.region,
        )
        self.assertIn(
            "Credential=my_key/{}/my_region/my_service/osc4_request".format(
                datetime.utcnow().strftime("%Y%m%d")
            ),
            headers["Authorization"],
        )

    def test_v4_signature_contains_signed_headers(self):
        action = "ReadImages"
        headers = self.signer.get_request_headers(
            action=action,
            data="{}",
            service_name=self.driver.service_name,
            region=self.driver.region,
        )
        self.assertIn("SignedHeaders=content-type;host;x-osc-date", headers["Authorization"])

    def test_get_signed_headers_contains_all_headers_lowercased(self):
        path = "my_region/my_service/outscale.com"
        headers = {
            "Content-Type": "application/json; charset=utf-8",
            "X-Osc-Date": self.now,
            "Host": path,
        }
        signed_headers = self.signer._get_signed_headers(headers)
        self.assertIn("content-type", signed_headers)
        self.assertIn("host", signed_headers)
        self.assertIn("x-osc-date", signed_headers)

    def test_get_signed_headers_concats_headers_sorted_lexically(self):
        path = "my_region/my_service/outscale.com"
        headers = {
            "Content-Type": "application/json; charset=utf-8",
            "X-Osc-Date": self.now,
            "Host": path,
        }
        signed_headers = self.signer._get_signed_headers(headers)
        self.assertEqual(signed_headers, "content-type;host;x-osc-date")

    def test_get_credential_scope(self):
        scope = self.signer._get_credential_scope(self.now)
        self.assertEqual(scope, "20150304/my_region/my_service/osc4_request")

    def test_get_canonical_headers_joins_all_headers(self):
        headers = {
            "Content-Type": "application/json; charset=utf-8",
            "X-Osc-Date": self.now,
            "Host": "my_region/my_service/outscale.com",
        }
        self.assertEqual(
            self.signer._get_canonical_headers(headers),
            "content-type:application/json; charset=utf-8\n"
            "host:my_region/my_service/outscale.com\n"
            "x-osc-date:2015-03-04 17:34:52\n",
        )

    def test_get_canonical_headers_sorts_headers_lexically(self):
        headers = {
            "accept-encoding": "gzip,deflate",
            "host": "my_host",
            "1st-header": "2",
            "x-osc-date": "20150304T173452Z",
            "user-agent": "my-ua",
        }
        self.assertEqual(
            self.signer._get_canonical_headers(headers),
            "1st-header:2\n"
            "accept-encoding:gzip,deflate\n"
            "host:my_host\n"
            "user-agent:my-ua\n"
            "x-osc-date:20150304T173452Z\n",
        )

    def test_get_canonical_headers_lowercases_headers_names(self):
        headers = {"Accept-Encoding": "GZIP,DEFLATE", "User-Agent": "My-UA"}
        self.assertEqual(
            self.signer._get_canonical_headers(headers),
            "accept-encoding:GZIP,DEFLATE\n" "user-agent:My-UA\n",
        )


if __name__ == "__main__":
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for Google Connection classes.
"""
import os
import sys
import time
import random
import urllib
import datetime
import unittest
import threading
from unittest import mock

import requests

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.utils.py3 import httplib
from libcloud.common.google import (
    GoogleAuthType,
    GoogleAuthError,
    GoogleBaseConnection,
    GoogleOAuth2Credential,
    GoogleBaseAuthConnection,
    GoogleServiceAcctAuthConnection,
    GoogleInstalledAppAuthConnection,
    GoogleGCEServiceAcctAuthConnection,
    _utcnow,
    _utc_timestamp,
)

try:
    import simplejson as json
except ImportError:
    import json


# Skip some tests if cryptography is unavailable
try:
    from cryptography.hazmat.primitives.hashes import SHA256
except ImportError:
    SHA256 = None


SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))
PEM_KEY = os.path.join(SCRIPT_PATH, "fixtures", "google", "pkey.pem")
JSON_KEY = os.path.join(SCRIPT_PATH, "fixtures", "google", "pkey.json")
JSON_KEY_INVALID = os.path.join(SCRIPT_PATH, "fixtures", "google", "pkey_invalid.json")
with open(JSON_KEY) as f:
    KEY_STR = json.loads(f.read())["private_key"]


GCE_PARAMS = ("email@developer.gserviceaccount.com", "key")
GCE_PARAMS_PEM_KEY = ("email@developer.gserviceaccount.com", PEM_KEY)
GCE_PARAMS_JSON_KEY = ("email@developer.gserviceaccount.com", JSON_KEY)
GCE_PARAMS_KEY = ("email@developer.gserviceaccount.com", KEY_STR)
GCE_PARAMS_IA = ("client_id", "client_secret")
GCE_PARAMS_IA_2 = ("client_id@test.apps.googleusercontent.com", "client_secret")
GCE_PARAMS_GCE = ("foo", "bar")
# GOOG + 16 alphanumeric chars
GCS_S3_PARAMS_20 = (
    "GOOG0123456789ABCXYZ",
    # 40 base64 chars
    "0102030405060708091011121314151617181920",
)
# GOOG + 20 alphanumeric chars
GCS_S3_PARAMS_24 = (
    "GOOGDF5OVRRGU4APFNSTVCXI",
    # 40 base64 chars
    "0102030405060708091011121314151617181920",
)
# GOOG + 57 alphanumeric chars
GCS_S3_PARAMS_61 = (
    "GOOGDF5OVRRGU4APFNSTVCXIRRGU4AP56789ABCX56789ABCXRRGU4APFNSTV",
    # 40 base64 chars
    "0102030405060708091011121314151617181920",
)
PEM_KEY_FILE = os.path.join(SCRIPT_PATH, "fixtures", "google", "pkey.pem")
PEM_KEY_FILE_INVALID = os.path.join(SCRIPT_PATH, "fixtures", "google", "pkey_invalid.pem")

JSON_KEY_FILE = os.path.join(SCRIPT_PATH, "fixtures", "google", "pkey.json")

with open(JSON_KEY_FILE) as f:
    PEM_KEY_STR = json.loads(f.read())["private_key"]

with open(JSON_KEY_FILE) as f:
    JSON_KEY_STR = f.read()
    JSON_KEY = json.loads(JSON_KEY_STR)


GCE_USERID_EMAIL = "email@developer.gserviceaccount.com"
GCE_PARAMS = (GCE_USERID_EMAIL, "key")
GCE_PARAMS_PEM_KEY_FILE = (GCE_USERID_EMAIL, PEM_KEY_FILE)
GCE_PARAMS_PEM_KEY_FILE_INVALID = (GCE_USERID_EMAIL, PEM_KEY_FILE_INVALID)
GCE_PARAMS_PEM_KEY = (GCE_USERID_EMAIL, PEM_KEY_STR)
GCE_PARAMS_JSON_KEY_FILE = (GCE_USERID_EMAIL, JSON_KEY_FILE)
GCE_PARAMS_JSON_KEY = (GCE_USERID_EMAIL, JSON_KEY)
GCE_PARAMS_JSON_KEY_INVALID = (GCE_USERID_EMAIL, JSON_KEY_INVALID)
GCE_PARAMS_JSON_KEY_STR = (GCE_USERID_EMAIL, JSON_KEY_STR)
GCE_PARAMS_IA = ("client_id", "client_secret")
GCE_PARAMS_GCE = ("foo", "bar")
GCS_S3_PARAMS_20 = (
    "GOOG0123456789ABCXYZ",  # GOOG + 16 alphanumeric chars
    "0102030405060708091011121314151617181920",
)  # 40 base64 chars
GCS_S3_PARAMS_24 = (
    "GOOGDF5OVRRGU4APFNSTVCXI",  # GOOG + 20 alphanumeric chars
    "0102030405060708091011121314151617181920",
)  # 40 base64 chars

STUB_UTCNOW = _utcnow()

STUB_TOKEN = {"access_token": "tokentoken", "token_type": "Bearer", "expires_in": 3600}

STUB_IA_TOKEN = {
    "access_token": "installedapp",
    "token_type": "Bearer",
    "expires_in": 3600,
    "refresh_token": "refreshrefresh",
}

STUB_REFRESH_TOKEN = {
    "access_token": "refreshrefresh",
    "token_type": "Bearer",
    "expires_in": 3600,
}

STUB_TOKEN_FROM_FILE = {
    "access_token": "token_from_file",
    "token_type": "Bearer",
    "expire_time": _utc_timestamp(STUB_UTCNOW + datetime.timedelta(seconds=3600)),
    "expires_in": 3600,
}


class MockJsonResponse:
    def __init__(self, body):
        self.object = body


class GoogleTestCase(LibcloudTestCase):
    """
    Assists in making Google tests hermetic and deterministic.

    Add anything that needs to be mocked here. Create a patcher with the
    suffix '_patcher'.

    e.g.
        _foo_patcher = mock.patch('module.submodule.class.foo', ...)

    Patchers are started at setUpClass and stopped at tearDownClass.

    Ideally, you should make a note in the thing being mocked, for clarity.
    """

    PATCHER_SUFFIX = "_patcher"

    _utcnow_patcher = mock.patch("libcloud.common.google._utcnow", return_value=STUB_UTCNOW)

    _authtype_is_gce_patcher = mock.patch(
        "libcloud.common.google.GoogleAuthType._is_gce", return_value=False
    )

    _read_token_file_patcher = mock.patch(
        "libcloud.common.google.GoogleOAuth2Credential._get_token_from_file",
        return_value=STUB_TOKEN_FROM_FILE,
    )

    _write_token_file_patcher = mock.patch(
        "libcloud.common.google.GoogleOAuth2Credential._write_token_to_file"
    )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        for patcher in [a for a in dir(cls) if a.endswith(cls.PATCHER_SUFFIX)]:
            getattr(cls, patcher).start()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        for patcher in [a for a in dir(cls) if a.endswith(cls.PATCHER_SUFFIX)]:
            getattr(cls, patcher).stop()


class GoogleBaseAuthConnectionTest(GoogleTestCase):
    """
    Tests for GoogleBaseAuthConnection
    """

    def setUp(self):
        GoogleBaseAuthConnection.conn_class = GoogleAuthMockHttp
        self.mock_scopes = ["foo", "bar"]
        kwargs = {"scopes": self.mock_scopes}
        self.conn = GoogleInstalledAppAuthConnection(*GCE_PARAMS, **kwargs)

    def test_scopes(self):
        self.assertEqual(self.conn.scopes, "foo bar")

    def test_add_default_headers(self):
        old_headers = {}
        expected_headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Host": "accounts.google.com",
        }
        new_headers = self.conn.add_default_headers(old_headers)
        self.assertEqual(new_headers, expected_headers)

    def test_token_request(self):
        request_body = {
            "code": "asdf",
            "client_id": self.conn.user_id,
            "client_secret": self.conn.key,
            "redirect_uri": self.conn.redirect_uri,
            "grant_type": "authorization_code",
        }
        new_token = self.conn._token_request(request_body)
        self.assertEqual(new_token["access_token"], STUB_IA_TOKEN["access_token"])
        exp = STUB_UTCNOW + datetime.timedelta(seconds=STUB_IA_TOKEN["expires_in"])
        self.assertEqual(new_token["expire_time"], _utc_timestamp(exp))


class GoogleInstalledAppAuthConnectionTest(GoogleTestCase):
    """
    Tests for GoogleInstalledAppAuthConnection
    """

    _ia_get_code_patcher = mock.patch(
        "libcloud.common.google.GoogleInstalledAppAuthConnection.get_code",
        return_value=1234,
    )

    def setUp(self):
        GoogleInstalledAppAuthConnection.conn_class = GoogleAuthMockHttp
        self.mock_scopes = ["https://www.googleapis.com/auth/foo"]
        kwargs = {"scopes": self.mock_scopes}
        self.conn = GoogleInstalledAppAuthConnection(*GCE_PARAMS, **kwargs)

    def test_refresh_token(self):
        # This token info doesn't have a refresh token, so a new token will be
        # requested
        token_info1 = {
            "access_token": "tokentoken",
            "token_type": "Bearer",
            "expires_in": 3600,
        }
        new_token1 = self.conn.refresh_token(token_info1)
        self.assertEqual(new_token1["access_token"], STUB_IA_TOKEN["access_token"])

        # This token info has a refresh token, so it will be able to be
        # refreshed.
        token_info2 = {
            "access_token": "tokentoken",
            "token_type": "Bearer",
            "expires_in": 3600,
            "refresh_token": "refreshrefresh",
        }
        new_token2 = self.conn.refresh_token(token_info2)
        self.assertEqual(new_token2["access_token"], STUB_REFRESH_TOKEN["access_token"])

        # Both sets should have refresh info
        self.assertTrue("refresh_token" in new_token1)
        self.assertTrue("refresh_token" in new_token2)


class GoogleInstalledAppAuthConnectionFirstLoginTest(LibcloudTestCase):
    def setUp(self):
        GoogleInstalledAppAuthConnection.conn_class = GoogleAuthMockHttp
        self.mock_scopes = ["https://www.googleapis.com/auth/foo"]
        kwargs = {"scopes": self.mock_scopes}
        self.conn = GoogleInstalledAppAuthConnection(*GCE_PARAMS, **kwargs)
        # We select random port on which HTTP server binds to, to avoid race conditions when
        # running multiple tests in parallel
        self.conn.redirect_uri_port = random.randint(5000, 20000)

    def test_it_receives_the_code_that_google_sends_via_local_loopback(self):
        expected_code = "1234ABC"
        received_code = self._do_first_sign_in(expected_code=expected_code, state=self.conn._state)
        self.assertEqual(received_code, expected_code)

    def test_it_aborts_if_state_is_suspicious(self):
        received_code = self._do_first_sign_in(
            expected_code="1234ABC", state=self.conn._state + "very suspicious"
        )
        self.assertEqual(received_code, None)

    def _do_first_sign_in(self, expected_code, state):
        """
        :param expected_code: The code that the fake Google sign-in local GET request will have in its query.
        :type expected_code: `str`
        :param state: The state that the fake Google sign-in local GET request will have in its query.
        :type state: `str`
        :return: The code that was extracted through local loopback.
        :rtype: `Optional[str]`
        """
        received_code = None

        def _get_code():
            nonlocal received_code
            received_code = self.conn.get_code()

        def _send_code():
            target_url = self.conn._redirect_uri_with_port
            params = {"state": state, "code": expected_code}
            params = urllib.parse.urlencode(params, quote_via=urllib.parse.quote)
            requests.get(url=target_url, params=params)

        fake_sign_in_thread = threading.Thread(target=_get_code)
        fake_google_response = threading.Thread(target=_send_code)

        fake_sign_in_thread.start()
        # TODO: Do a proper fix for race condition and only start get_code once _send_code
        # is initialized and up and running
        time.sleep(0.2)
        fake_google_response.start()
        fake_google_response.join()
        fake_sign_in_thread.join()

        return received_code


class GoogleAuthTypeTest(GoogleTestCase):
    def test_guess(self):
        self.assertEqual(GoogleAuthType.guess_type(GCE_PARAMS_IA[0]), GoogleAuthType.IA)
        with mock.patch.object(GoogleAuthType, "_is_gce", return_value=True):
            # Since _is_gce currently depends on the environment, not on
            # parameters, other auths should override GCE. It does not make
            # sense for IA auth to happen on GCE, which is why it's left out.
            self.assertEqual(GoogleAuthType.guess_type(GCE_PARAMS[0]), GoogleAuthType.SA)
            self.assertEqual(GoogleAuthType.guess_type(GCS_S3_PARAMS_20[0]), GoogleAuthType.GCS_S3)
            self.assertEqual(GoogleAuthType.guess_type(GCS_S3_PARAMS_24[0]), GoogleAuthType.GCS_S3)
            self.assertEqual(GoogleAuthType.guess_type(GCS_S3_PARAMS_61[0]), GoogleAuthType.GCS_S3)
            self.assertEqual(GoogleAuthType.guess_type(GCE_PARAMS_GCE[0]), GoogleAuthType.GCE)

    def test_guess_gce_metadata_server_not_called_for_ia(self):
        # Verify that we don't try to contact GCE metadata server in case IA
        # credentials are used
        with mock.patch.object(GoogleAuthType, "_is_gce", return_value=False):
            self.assertEqual(GoogleAuthType._is_gce.call_count, 0)
            self.assertEqual(GoogleAuthType.guess_type(GCE_PARAMS_IA_2[0]), GoogleAuthType.IA)
            self.assertEqual(GoogleAuthType._is_gce.call_count, 0)


class GoogleOAuth2CredentialTest(GoogleTestCase):
    _ia_get_code_patcher = mock.patch(
        "libcloud.common.google.GoogleInstalledAppAuthConnection.get_code",
        return_value=1234,
    )

    def test_init_oauth2(self):
        kwargs = {"auth_type": GoogleAuthType.IA}
        cred = GoogleOAuth2Credential(*GCE_PARAMS, **kwargs)

        # If there is a viable token file, this gets used first
        self.assertEqual(cred.token, STUB_TOKEN_FROM_FILE)

        # No token file, get a new token. Check that it gets written to file.
        with mock.patch.object(GoogleOAuth2Credential, "_get_token_from_file", return_value=None):
            cred = GoogleOAuth2Credential(*GCE_PARAMS, **kwargs)
            expected = STUB_IA_TOKEN
            expected["expire_time"] = cred.token["expire_time"]
            self.assertEqual(cred.token, expected)
            cred._write_token_to_file.assert_called_once_with()

    def test_refresh(self):
        args = list(GCE_PARAMS) + [GoogleAuthType.GCE]
        cred = GoogleOAuth2Credential(*args)
        cred._refresh_token = mock.Mock()

        # Test getting an unexpired access token.
        tomorrow = datetime.datetime.now() + datetime.timedelta(days=1)
        cred.token = {
            "access_token": "Access Token!",
            "expire_time": _utc_timestamp(tomorrow),
        }
        cred.access_token
        self.assertFalse(cred._refresh_token.called)

        # Test getting an expired access token.
        yesterday = datetime.datetime.now() - datetime.timedelta(days=1)
        cred.token = {
            "access_token": "Access Token!",
            "expire_time": _utc_timestamp(yesterday),
        }
        cred.access_token
        self.assertTrue(cred._refresh_token.called)

    def test_auth_connection(self):
        # Test a bogus auth type
        self.assertRaises(
            GoogleAuthError, GoogleOAuth2Credential, *GCE_PARAMS, **{"auth_type": "XX"}
        )
        # Try to create an OAuth2 credential when dealing with a GCS S3
        # interoperability auth type.
        self.assertRaises(
            GoogleAuthError,
            GoogleOAuth2Credential,
            *GCE_PARAMS,
            **{"auth_type": GoogleAuthType.GCS_S3},
        )

        kwargs = {}

        if SHA256:
            kwargs["auth_type"] = GoogleAuthType.SA
            cred1 = GoogleOAuth2Credential(*GCE_PARAMS_PEM_KEY_FILE, **kwargs)
            self.assertTrue(isinstance(cred1.oauth2_conn, GoogleServiceAcctAuthConnection))

            cred1 = GoogleOAuth2Credential(*GCE_PARAMS_JSON_KEY_FILE, **kwargs)
            self.assertTrue(isinstance(cred1.oauth2_conn, GoogleServiceAcctAuthConnection))

            cred1 = GoogleOAuth2Credential(*GCE_PARAMS_PEM_KEY, **kwargs)
            self.assertTrue(isinstance(cred1.oauth2_conn, GoogleServiceAcctAuthConnection))

            cred1 = GoogleOAuth2Credential(*GCE_PARAMS_JSON_KEY, **kwargs)
            self.assertTrue(isinstance(cred1.oauth2_conn, GoogleServiceAcctAuthConnection))

            cred1 = GoogleOAuth2Credential(*GCE_PARAMS_KEY, **kwargs)
            self.assertTrue(isinstance(cred1.oauth2_conn, GoogleServiceAcctAuthConnection))

            kwargs["auth_type"] = GoogleAuthType.SA
            cred1 = GoogleOAuth2Credential(*GCE_PARAMS_JSON_KEY_STR, **kwargs)
            self.assertTrue(isinstance(cred1.oauth2_conn, GoogleServiceAcctAuthConnection))

            self.assertRaises(GoogleAuthError, GoogleOAuth2Credential, *GCE_PARAMS, **kwargs)

            # Invalid pem key
            kwargs["auth_type"] = GoogleAuthType.SA
            expected_msg = "Unable to decode provided PEM key:"
            self.assertRaisesRegex(
                GoogleAuthError,
                expected_msg,
                GoogleOAuth2Credential,
                *GCE_PARAMS_PEM_KEY_FILE_INVALID,
                **kwargs,
            )

            kwargs["auth_type"] = GoogleAuthType.SA
            expected_msg = "Unable to decode provided PEM key:"
            self.assertRaisesRegex(
                GoogleAuthError,
                expected_msg,
                GoogleOAuth2Credential,
                *GCE_PARAMS_JSON_KEY_INVALID,
                **kwargs,
            )

        kwargs["auth_type"] = GoogleAuthType.IA
        cred2 = GoogleOAuth2Credential(*GCE_PARAMS_IA, **kwargs)
        self.assertTrue(isinstance(cred2.oauth2_conn, GoogleInstalledAppAuthConnection))

        kwargs["auth_type"] = GoogleAuthType.GCE
        cred3 = GoogleOAuth2Credential(*GCE_PARAMS_GCE, **kwargs)
        self.assertTrue(isinstance(cred3.oauth2_conn, GoogleGCEServiceAcctAuthConnection))


class GoogleBaseConnectionTest(GoogleTestCase):
    """
    Tests for GoogleBaseConnection
    """

    def setUp(self):
        GoogleBaseAuthConnection.conn_class = GoogleAuthMockHttp
        self.mock_scopes = ["https://www.googleapis.com/auth/foo"]
        kwargs = {"scopes": self.mock_scopes, "auth_type": GoogleAuthType.IA}
        self.conn = GoogleBaseConnection(*GCE_PARAMS, **kwargs)

    def test_add_default_headers(self):
        old_headers = {}
        new_expected_headers = {
            "Content-Type": "application/json",
            "Host": "www.googleapis.com",
        }
        new_headers = self.conn.add_default_headers(old_headers)
        self.assertEqual(new_headers, new_expected_headers)

    def test_pre_connect_hook(self):
        old_params = {}
        old_headers = {}
        auth_str = "{} {}".format(
            STUB_TOKEN_FROM_FILE["token_type"],
            STUB_TOKEN_FROM_FILE["access_token"],
        )
        new_expected_params = {}
        new_expected_headers = {"Authorization": auth_str}
        new_params, new_headers = self.conn.pre_connect_hook(old_params, old_headers)
        self.assertEqual(new_params, new_expected_params)
        self.assertEqual(new_headers, new_expected_headers)

    def test_encode_data(self):
        data = {"key": "value"}
        json_data = '{"key": "value"}'
        encoded_data = self.conn.encode_data(data)
        self.assertEqual(encoded_data, json_data)

    def test_has_completed(self):
        body1 = {
            "endTime": "2013-06-26T10:05:07.630-07:00",
            "id": "3681664092089171723",
            "kind": "compute#operation",
            "status": "DONE",
            "targetId": "16211908079305042870",
        }
        body2 = {
            "endTime": "2013-06-26T10:05:07.630-07:00",
            "id": "3681664092089171723",
            "kind": "compute#operation",
            "status": "RUNNING",
            "targetId": "16211908079305042870",
        }
        response1 = MockJsonResponse(body1)
        response2 = MockJsonResponse(body2)
        self.assertTrue(self.conn.has_completed(response1))
        self.assertFalse(self.conn.has_completed(response2))

    def test_get_poll_request_kwargs(self):
        body = {
            "endTime": "2013-06-26T10:05:07.630-07:00",
            "id": "3681664092089171723",
            "kind": "compute#operation",
            "selfLink": "https://www.googleapis.com/operations-test",
        }
        response = MockJsonResponse(body)
        expected_kwargs = {"action": "https://www.googleapis.com/operations-test"}
        kwargs = self.conn.get_poll_request_kwargs(response, None, {})
        self.assertEqual(kwargs, expected_kwargs)

    def test_morph_action_hook(self):
        self.conn.request_path = "/compute/apiver/project/project-name"
        action1 = "https://www.googleapis.com/compute/apiver/project" "/project-name/instances"
        action2 = "/instances"
        expected_request = "/compute/apiver/project/project-name/instances"
        request1 = self.conn.morph_action_hook(action1)
        request2 = self.conn.morph_action_hook(action2)
        self.assertEqual(request1, expected_request)
        self.assertEqual(request2, expected_request)


class GoogleAuthMockHttp(MockHttp):
    """
    Mock HTTP Class for Google Auth Connections.
    """

    json_hdr = {"content-type": "application/json; charset=UTF-8"}

    def _o_oauth2_token(self, method, url, body, headers):
        if "code" in body:
            body = json.dumps(STUB_IA_TOKEN)
        elif "refresh_token" in body:
            body = json.dumps(STUB_REFRESH_TOKEN)
        else:
            body = json.dumps(STUB_TOKEN)
        return (httplib.OK, body, self.json_hdr, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import json
from unittest.mock import Mock, call

from libcloud.test import unittest
from libcloud.compute.base import NodeSize, NodeImage, NodeLocation, NodeAuthSSHKey
from libcloud.common.upcloud import (
    PlanPrice,
    UpcloudNodeDestroyer,
    UpcloudNodeOperations,
    UpcloudTimeoutException,
    UpcloudCreateNodeRequestBody,
    _StorageDevice,
)


class TestUpcloudCreateNodeRequestBody(unittest.TestCase):
    def setUp(self):
        self.image = NodeImage(
            id="01000000-0000-4000-8000-000030060200",
            name="Ubuntu Server 16.04 LTS (Xenial Xerus)",
            driver="",
            extra={"type": "template"},
        )
        self.location = NodeLocation(id="fi-hel1", name="Helsinki #1", country="FI", driver="")
        self.size = NodeSize(
            id="1xCPU-1GB",
            name="1xCPU-1GB",
            ram=1024,
            disk=30,
            bandwidth=2048,
            extra={"core_number": 1, "storage_tier": "maxiops"},
            price=None,
            driver="",
        )

    def test_creating_node_from_template_image(self):
        body = UpcloudCreateNodeRequestBody(
            name="ts", image=self.image, location=self.location, size=self.size
        )
        json_body = body.to_json()
        dict_body = json.loads(json_body)
        expected_body = {
            "server": {
                "title": "ts",
                "hostname": "localhost",
                "plan": "1xCPU-1GB",
                "zone": "fi-hel1",
                "login_user": {"username": "root", "create_password": "yes"},
                "storage_devices": {
                    "storage_device": [
                        {
                            "action": "clone",
                            "title": "Ubuntu Server 16.04 LTS (Xenial Xerus)",
                            "storage": "01000000-0000-4000-8000-000030060200",
                            "size": 30,
                            "tier": "maxiops",
                        }
                    ]
                },
            }
        }
        self.assertDictEqual(expected_body, dict_body)

    def test_creating_node_from_cdrom_image(self):
        image = NodeImage(
            id="01000000-0000-4000-8000-000030060200",
            name="Ubuntu Server 16.04 LTS (Xenial Xerus)",
            driver="",
            extra={"type": "cdrom"},
        )
        body = UpcloudCreateNodeRequestBody(
            name="ts", image=image, location=self.location, size=self.size
        )
        json_body = body.to_json()
        dict_body = json.loads(json_body)
        expected_body = {
            "server": {
                "title": "ts",
                "hostname": "localhost",
                "plan": "1xCPU-1GB",
                "zone": "fi-hel1",
                "login_user": {"username": "root", "create_password": "yes"},
                "storage_devices": {
                    "storage_device": [
                        {
                            "action": "create",
                            "size": 30,
                            "tier": "maxiops",
                            "title": "Ubuntu Server 16.04 LTS (Xenial Xerus)",
                        },
                        {
                            "action": "attach",
                            "storage": "01000000-0000-4000-8000-000030060200",
                            "type": "cdrom",
                        },
                    ]
                },
            }
        }
        self.assertDictEqual(expected_body, dict_body)

    def test_creating_node_using_ssh_keys(self):
        auth = NodeAuthSSHKey("sshkey")

        body = UpcloudCreateNodeRequestBody(
            name="ts",
            image=self.image,
            location=self.location,
            size=self.size,
            auth=auth,
        )
        json_body = body.to_json()
        dict_body = json.loads(json_body)
        expected_body = {
            "server": {
                "title": "ts",
                "hostname": "localhost",
                "plan": "1xCPU-1GB",
                "zone": "fi-hel1",
                "login_user": {
                    "username": "root",
                    "ssh_keys": {"ssh_key": ["sshkey"]},
                },
                "storage_devices": {
                    "storage_device": [
                        {
                            "action": "clone",
                            "size": 30,
                            "title": "Ubuntu Server 16.04 LTS (Xenial Xerus)",
                            "tier": "maxiops",
                            "storage": "01000000-0000-4000-8000-000030060200",
                        }
                    ]
                },
            }
        }
        self.assertDictEqual(expected_body, dict_body)

    def test_creating_node_using_hostname(self):
        body = UpcloudCreateNodeRequestBody(
            name="ts",
            image=self.image,
            location=self.location,
            size=self.size,
            ex_hostname="myhost.upcloud.com",
        )
        json_body = body.to_json()
        dict_body = json.loads(json_body)
        expected_body = {
            "server": {
                "title": "ts",
                "hostname": "myhost.upcloud.com",
                "plan": "1xCPU-1GB",
                "zone": "fi-hel1",
                "login_user": {"username": "root", "create_password": "yes"},
                "storage_devices": {
                    "storage_device": [
                        {
                            "action": "clone",
                            "title": "Ubuntu Server 16.04 LTS (Xenial Xerus)",
                            "storage": "01000000-0000-4000-8000-000030060200",
                            "tier": "maxiops",
                            "size": 30,
                        }
                    ]
                },
            }
        }
        self.assertDictEqual(expected_body, dict_body)

    def test_creating_node_with_non_default_username(self):
        body = UpcloudCreateNodeRequestBody(
            name="ts",
            image=self.image,
            location=self.location,
            size=self.size,
            ex_username="someone",
        )
        json_body = body.to_json()
        dict_body = json.loads(json_body)

        login_user = dict_body["server"]["login_user"]
        self.assertDictEqual({"username": "someone", "create_password": "yes"}, login_user)


class TestStorageDevice(unittest.TestCase):
    def setUp(self):
        self.image = NodeImage(
            id="01000000-0000-4000-8000-000030060200",
            name="Ubuntu Server 16.04 LTS (Xenial Xerus)",
            driver="",
            extra={"type": "template"},
        )
        self.size = NodeSize(
            id="1xCPU-1GB",
            name="1xCPU-1GB",
            ram=1024,
            disk=30,
            bandwidth=2048,
            extra={"core_number": 1},
            price=None,
            driver="",
        )

    def test_storage_tier_default_value(self):
        storagedevice = _StorageDevice(self.image, self.size)
        d = storagedevice.to_dict()

        self.assertEqual(d["storage_device"][0]["tier"], "maxiops")

    def test_storage_tier_given(self):
        self.size.extra["storage_tier"] = "hdd"
        storagedevice = _StorageDevice(self.image, self.size)
        d = storagedevice.to_dict()

        self.assertEqual(d["storage_device"][0]["tier"], "hdd")


class TestUpcloudNodeDestroyer(unittest.TestCase):
    def setUp(self):
        self.mock_sleep = Mock()
        self.mock_operations = Mock(spec=UpcloudNodeOperations)
        self.destroyer = UpcloudNodeDestroyer(self.mock_operations, sleep_func=self.mock_sleep)

    def test_node_already_in_stopped_state(self):
        self.mock_operations.get_node_state.side_effect = ["stopped"]

        self.assertTrue(self.destroyer.destroy_node(1))

        self.assertTrue(self.mock_operations.stop_node.call_count == 0)
        self.mock_operations.destroy_node.assert_called_once_with(1)

    def test_node_in_error_state(self):
        self.mock_operations.get_node_state.side_effect = ["error"]

        self.assertFalse(self.destroyer.destroy_node(1))

        self.assertTrue(self.mock_operations.stop_node.call_count == 0)
        self.assertTrue(self.mock_operations.destroy_node.call_count == 0)

    def test_node_in_started_state(self):
        self.mock_operations.get_node_state.side_effect = ["started", "stopped"]

        self.assertTrue(self.destroyer.destroy_node(1))

        self.mock_operations.stop_node.assert_called_once_with(1)
        self.mock_operations.destroy_node.assert_called_once_with(1)

    def test_node_in_maintenace_state(self):
        self.mock_operations.get_node_state.side_effect = [
            "maintenance",
            "maintenance",
            None,
        ]

        self.assertTrue(self.destroyer.destroy_node(1))

        self.mock_sleep.assert_has_calls(
            [call(self.destroyer.WAIT_AMOUNT), call(self.destroyer.WAIT_AMOUNT)]
        )

        self.assertTrue(self.mock_operations.stop_node.call_count == 0)
        self.assertTrue(self.mock_operations.destroy_node.call_count == 0)

    def test_node_statys_in_started_state_for_awhile(self):
        self.mock_operations.get_node_state.side_effect = [
            "started",
            "started",
            "stopped",
        ]

        self.assertTrue(self.destroyer.destroy_node(1))

        # Only one all for stop should be done
        self.mock_operations.stop_node.assert_called_once_with(1)
        self.mock_sleep.assert_has_calls([call(self.destroyer.WAIT_AMOUNT)])
        self.mock_operations.destroy_node.assert_called_once_with(1)

    def test_reuse(self):
        """Verify that internal flag self.destroyer._stop_node is handled properly"""
        self.mock_operations.get_node_state.side_effect = [
            "started",
            "stopped",
            "started",
            "stopped",
        ]
        self.assertTrue(self.destroyer.destroy_node(1))
        self.assertTrue(self.destroyer.destroy_node(1))

        self.assertEqual(self.mock_sleep.call_count, 0)
        self.assertEqual(self.mock_operations.stop_node.call_count, 2)

    def test_timeout(self):
        self.mock_operations.get_node_state.side_effect = ["maintenance"] * 50

        self.assertRaises(UpcloudTimeoutException, self.destroyer.destroy_node, 1)

    def test_timeout_reuse(self):
        """Verify sleep count is handled properly"""
        self.mock_operations.get_node_state.side_effect = ["maintenance"] * 50
        self.assertRaises(UpcloudTimeoutException, self.destroyer.destroy_node, 1)

        self.mock_operations.get_node_state.side_effect = ["maintenance", None]
        self.assertTrue(self.destroyer.destroy_node(1))


class TestPlanPrice(unittest.TestCase):
    def setUp(self):
        prices = [
            {"name": "uk-lon1", "server_plan_1xCPU-1GB": {"amount": 1, "price": 1.488}},
            {"name": "fi-hel1", "server_plan_1xCPU-1GB": {"amount": 1, "price": 1.588}},
        ]
        self.pp = PlanPrice(prices)

    def test_zone_prices(self):
        location = NodeLocation(id="fi-hel1", name="Helsinki #1", country="FI", driver=None)
        self.assertEqual(self.pp.get_price("1xCPU-1GB", location), 1.588)

    def test_plan_not_found_in_zone(self):
        location = NodeLocation(id="no_such_location", name="", country="", driver=None)
        self.assertIsNone(self.pp.get_price("1xCPU-1GB", location))

    def test_no_location_given(self):
        self.assertIsNone(self.pp.get_price("1xCPU-1GB"))


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.base import Zone, Record
from libcloud.dns.types import (
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
    RecordAlreadyExistsError,
)
from libcloud.utils.py3 import httplib
from libcloud.common.nsone import NsOneException
from libcloud.test.secrets import DNS_PARAMS_NSONE
from libcloud.dns.drivers.nsone import NsOneDNSDriver
from libcloud.test.file_fixtures import DNSFileFixtures


class NsOneTests(unittest.TestCase):
    def setUp(self):
        NsOneMockHttp.type = None
        NsOneDNSDriver.connectionCls.conn_class = NsOneMockHttp
        self.driver = NsOneDNSDriver(*DNS_PARAMS_NSONE)
        self.test_zone = Zone(
            id="test.com",
            type="master",
            ttl=None,
            domain="test.com",
            extra={},
            driver=self,
        )
        self.test_record = Record(
            id="13",
            type=RecordType.A,
            name="example.com",
            zone=self.test_zone,
            data="127.0.0.1",
            driver=self,
            extra={},
        )

    def test_list_zones_empty(self):
        NsOneMockHttp.type = "EMPTY_ZONES_LIST"
        zones = self.driver.list_zones()

        self.assertEqual(zones, [])

    def test_list_zones_success(self):
        zones = self.driver.list_zones()

        self.assertEqual(len(zones), 2)

        zone = zones[0]
        self.assertEqual(zone.id, "520422af9f782d37dffb588b")
        self.assertIsNone(zone.type)
        self.assertEqual(zone.domain, "example.com")
        self.assertEqual(zone.ttl, 3600)

        zone = zones[1]
        self.assertEqual(zone.id, "520422c99f782d37dffb5892")
        self.assertIsNone(zone.type)
        self.assertEqual(zone.domain, "nsoneisgreat.com")
        self.assertEqual(zone.ttl, 3600)

    def test_delete_zone_zone_does_not_exist(self):
        NsOneMockHttp.type = "DELETE_ZONE_ZONE_DOES_NOT_EXIST"

        try:
            self.driver.delete_zone(zone=self.test_zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, self.test_zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_delete_zone_success(self):
        NsOneMockHttp.type = "DELETE_ZONE_SUCCESS"
        status = self.driver.delete_zone(zone=self.test_zone)

        self.assertTrue(status)

    def test_get_zone_zone_does_not_exist(self):
        NsOneMockHttp.type = "GET_ZONE_ZONE_DOES_NOT_EXIST"
        try:
            self.driver.get_zone(zone_id="zonedoesnotexist.com")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "zonedoesnotexist.com")
        else:
            self.fail("Exception was not thrown")

    def test_create_zone_success(self):
        NsOneMockHttp.type = "CREATE_ZONE_SUCCESS"
        zone = self.driver.create_zone(domain="newzone.com")

        self.assertEqual(zone.id, "52051b2c9f782d58bb4df41b")
        self.assertEqual(zone.domain, "newzone.com")
        self.assertIsNone(zone.type),
        self.assertEqual(zone.ttl, 3600)

    def test_create_zone_zone_already_exists(self):
        NsOneMockHttp.type = "CREATE_ZONE_ZONE_ALREADY_EXISTS"

        try:
            self.driver.create_zone(domain="newzone.com")
        except ZoneAlreadyExistsError as e:
            self.assertEqual(e.zone_id, "newzone.com")
        else:
            self.fail("Exception was not thrown")

    def test_get_record_record_does_not_exist(self):
        NsOneMockHttp.type = "GET_RECORD_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="getrecord.com", record_id="A")
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, "A")
        else:
            self.fail("Exception was not thrown")

    def test_get_record_success(self):
        NsOneMockHttp.type = "GET_RECORD_SUCCESS"
        record = self.driver.get_record(zone_id="getrecord.com", record_id="A")

        self.assertEqual(record.id, "520519509f782d58bb4df419")
        self.assertEqual(record.name, "www.example.com")
        self.assertEqual(record.data, ["1.1.1.1"])
        self.assertEqual(record.type, RecordType.A)

    def test_list_records_zone_does_not_exist(self):
        NsOneMockHttp.type = "LIST_RECORDS_ZONE_DOES_NOT_EXIST"

        try:
            self.driver.list_records(zone=self.test_zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, self.test_zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_list_records_empty(self):
        NsOneMockHttp.type = "LIST_RECORDS_EMPTY"
        records = self.driver.list_records(zone=self.test_zone)

        self.assertEqual(records, [])

    def test_list_records_success(self):
        NsOneMockHttp.type = "LIST_RECORDS_SUCCESS"
        records = self.driver.list_records(zone=self.test_zone)
        self.assertEqual(len(records), 2)

        arecord = records[1]
        self.assertEqual(arecord.id, "520519509f782d58bb4df419")
        self.assertEqual(arecord.name, "www.example.com")
        self.assertEqual(arecord.type, RecordType.A)
        self.assertEqual(arecord.data, ["1.2.3.4"])

    def test_create_record_success(self):
        NsOneMockHttp.type = "CREATE_RECORD_SUCCESS"
        arecord = self.driver.create_record(
            self.test_record.name,
            self.test_record.zone,
            self.test_record.type,
            self.test_record.data,
            self.test_record.extra,
        )
        self.assertEqual(arecord.id, "608f9619ebe68600ac9f807d")
        self.assertEqual(arecord.name, "test.com")
        self.assertEqual(arecord.type, RecordType.A)
        self.assertEqual(arecord.data, ["127.0.0.1"])

    def test_create_record_already_exists(self):
        NsOneMockHttp.type = "CREATE_RECORD_ALREADY_EXISTS"
        try:
            self.driver.create_record(
                self.test_record.name,
                self.test_record.zone,
                self.test_record.type,
                self.test_record.data,
                self.test_record.extra,
            )
        except RecordAlreadyExistsError as err:
            self.assertEqual(err.value, "record already exists")
        else:
            self.fail("Exception was not thrown")

    def test_create_record_zone_not_found(self):
        NsOneMockHttp.type = "CREATE_RECORD_ZONE_NOT_FOUND"
        try:
            self.driver.create_record(
                self.test_record.name,
                self.test_record.zone,
                self.test_record.type,
                self.test_record.data,
                self.test_record.extra,
            )
        except NsOneException as err:
            self.assertEqual(err.message, "zone not found")
        else:
            self.fail("Exception was not thrown")

    def test_delete_record_record_does_not_exist(self):
        NsOneMockHttp.type = "DELETE_RECORD_RECORD_DOES_NOT_EXIST"

        try:
            self.driver.delete_record(record=self.test_record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, self.test_record.id)
        else:
            self.fail("Exception was not thrown")

    def test_delete_record_success(self):
        NsOneMockHttp.type = "DELETE_RECORD_SUCCESS"
        status = self.driver.delete_record(record=self.test_record)

        self.assertTrue(status)


class NsOneMockHttp(MockHttp):
    fixtures = DNSFileFixtures("nsone")

    def _v1_zones_EMPTY_ZONES_LIST(self, method, url, body, headers):
        body = self.fixtures.load("empty_zones_list.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones(self, method, url, body, headers):
        body = self.fixtures.load("list_zones.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_getzone_com_GET_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_zonedoesnotexist_com_GET_ZONE_ZONE_DOES_NOT_EXIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("zone_does_not_exist.json")

        return 404, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_DELETE_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("delete_zone_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_DELETE_ZONE_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("zone_does_not_exist.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_newzone_com_CREATE_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("create_zone_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_newzone_com_CREATE_ZONE_ZONE_ALREADY_EXISTS(self, method, url, body, headers):
        body = self.fixtures.load("zone_already_exists.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_LIST_RECORDS_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_LIST_RECORDS_EMPTY(self, method, url, body, headers):
        body = self.fixtures.load("list_records_empty.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_LIST_RECORDS_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("zone_does_not_exist.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_example_com_A_DELETE_RECORD_RECORD_DOES_NOT_EXIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("record_does_not_exist.json")

        return 404, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_example_com_A_DELETE_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("delete_record_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_getrecord_com_getrecord_com_A_GET_RECORD_SUCCESS(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("get_record_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_getrecord_com_GET_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_getrecord_com_getrecord_com_A_GET_RECORD_DOES_NOT_EXIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("record_does_not_exist.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_example_com_test_com_A_CREATE_RECORD_SUCCESS(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("create_record_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_example_com_test_com_A_CREATE_RECORD_ALREADY_EXISTS(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("create_record_already_exists.json")

        return 404, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_test_com_example_com_test_com_A_CREATE_RECORD_ZONE_NOT_FOUND(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("create_record_zone_not_found.json")

        return 404, body, {}, httplib.responses[httplib.OK]


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from libcloud.test import MockHttp, unittest
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import CONTAINER_PARAMS_ECS
from libcloud.container.base import Container, ContainerImage, ContainerCluster
from libcloud.test.file_fixtures import ContainerFileFixtures
from libcloud.container.drivers.ecs import ElasticContainerDriver
from libcloud.container.utils.docker import RegistryClient


class ElasticContainerDriverTestCase(unittest.TestCase):
    def setUp(self):
        ElasticContainerDriver.connectionCls.conn_class = ECSMockHttp
        ECSMockHttp.type = None
        ECSMockHttp.use_param = "a"
        ElasticContainerDriver.ecrConnectionClass.conn_class = ECSMockHttp

        self.driver = ElasticContainerDriver(*CONTAINER_PARAMS_ECS)

    def test_list_clusters(self):
        clusters = self.driver.list_clusters()
        self.assertEqual(len(clusters), 1)
        self.assertEqual(clusters[0].id, "arn:aws:ecs:us-east-1:012345678910:cluster/default")
        self.assertEqual(clusters[0].name, "default")

    def test_create_cluster(self):
        cluster = self.driver.create_cluster("my-cluster")
        self.assertEqual(cluster.name, "my-cluster")

    def test_destroy_cluster(self):
        self.assertTrue(
            self.driver.destroy_cluster(
                ContainerCluster(
                    id="arn:aws:ecs:us-east-1:012345678910:cluster/jim",
                    name="jim",
                    driver=self.driver,
                )
            )
        )

    def test_list_containers(self):
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 1)

    def test_list_containers_for_cluster(self):
        cluster = self.driver.list_clusters()[0]
        containers = self.driver.list_containers(cluster=cluster)
        self.assertEqual(len(containers), 1)

    def test_deploy_container(self):
        container = self.driver.deploy_container(
            name="jim",
            image=ContainerImage(
                id=None, name="mysql", path="mysql", version=None, driver=self.driver
            ),
        )
        self.assertEqual(
            container.id,
            "arn:aws:ecs:ap-southeast-2:647433528374:container/e443d10f-dea3-481e-8a1e-966b9ad4e498",
        )

    def test_get_container(self):
        container = self.driver.get_container(
            "arn:aws:ecs:us-east-1:012345678910:container/76c980a8-2454-4a9c-acc4-9eb103117273"
        )
        self.assertEqual(
            container.id,
            "arn:aws:ecs:ap-southeast-2:647433528374:container/d56d4e2c-9804-42a7-9f2a-6029cb50d4a2",
        )
        self.assertEqual(container.name, "simple-app")
        self.assertEqual(container.image.name, "simple-app")

    def test_start_container(self):
        container = self.driver.start_container(
            Container(
                id=None,
                name=None,
                image=None,
                state=None,
                ip_addresses=None,
                driver=self.driver,
                extra={"taskDefinitionArn": ""},
            )
        )
        self.assertFalse(container is None)

    def test_stop_container(self):
        container = self.driver.stop_container(
            Container(
                id=None,
                name=None,
                image=None,
                state=None,
                ip_addresses=None,
                driver=self.driver,
                extra={"taskArn": "12345", "taskDefinitionArn": "123556"},
            )
        )
        self.assertFalse(container is None)

    def test_restart_container(self):
        container = self.driver.restart_container(
            Container(
                id=None,
                name=None,
                image=None,
                state=None,
                ip_addresses=None,
                driver=self.driver,
                extra={"taskArn": "12345", "taskDefinitionArn": "123556"},
            )
        )
        self.assertFalse(container is None)

    def test_list_images(self):
        images = self.driver.list_images("my-images")
        self.assertEqual(len(images), 1)
        self.assertEqual(
            images[0].name, "647433528374.dkr.ecr.region.amazonaws.com/my-images:latest"
        )

    def test_ex_create_service(self):
        cluster = self.driver.list_clusters()[0]
        task_definition = self.driver.list_containers()[0].extra["taskDefinitionArn"]
        service = self.driver.ex_create_service(
            cluster=cluster, name="jim", task_definition=task_definition
        )
        self.assertEqual(service["serviceName"], "test")

    def test_ex_list_service_arns(self):
        arns = self.driver.ex_list_service_arns()
        self.assertEqual(len(arns), 2)

    def test_ex_describe_service(self):
        arn = self.driver.ex_list_service_arns()[0]
        service = self.driver.ex_describe_service(arn)
        self.assertEqual(service["serviceName"], "test")

    def test_ex_destroy_service(self):
        arn = self.driver.ex_list_service_arns()[0]
        service = self.driver.ex_destroy_service(arn)
        self.assertEqual(service["status"], "DRAINING")

    def test_ex_get_registry_client(self):
        client = self.driver.ex_get_registry_client("my-images")
        self.assertIsInstance(client, RegistryClient)


class ECSMockHttp(MockHttp):
    fixtures = ContainerFileFixtures("ecs")
    fixture_map = {
        "DescribeClusters": "describeclusters.json",
        "CreateCluster": "createcluster.json",
        "DeleteCluster": "deletecluster.json",
        "DescribeTasks": "describetasks.json",
        "ListTasks": "listtasks.json",
        "ListClusters": "listclusters.json",
        "RegisterTaskDefinition": "registertaskdefinition.json",
        "RunTask": "runtask.json",
        "StopTask": "stoptask.json",
        "ListImages": "listimages.json",
        "DescribeRepositories": "describerepositories.json",
        "CreateService": "createservice.json",
        "ListServices": "listservices.json",
        "DescribeServices": "describeservices.json",
        "DeleteService": "deleteservice.json",
        "GetAuthorizationToken": "getauthorizationtoken.json",
    }

    def root(self, method, url, body, headers):
        target = headers["x-amz-target"]

        # Workaround for host not being correctly set for the tests
        if "%s" in self.host:
            self.host = self.host % ("region")

        if target is not None:
            type = target.split(".")[-1]
            if type is None or self.fixture_map.get(type) is None:
                raise AssertionError("Unsupported request type %s" % (target))
            body = self.fixtures.load(self.fixture_map.get(type))
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.storage.drivers.s3 import (
    S3_CDN_URL_EXPIRY_HOURS,
    S3StorageDriver,
    BaseS3StorageDriver,
    S3SignatureV4Connection,
)

__all__ = ["ScalewayStorageDriver"]

SCW_FR_PAR_STANDARD_HOST = "s3.fr-par.scw.cloud"
SCW_NL_AMS_HOST = "s3.nl-ams.scw.cloud"
SCW_PL_WAR_HOST = "s3.pl-waw.scw.cloud"

# Maps SCW region name to connection hostname
REGION_TO_HOST_MAP = {
    "fr-par": SCW_FR_PAR_STANDARD_HOST,
    "nl-ams": SCW_NL_AMS_HOST,
    "pl-waw": SCW_PL_WAR_HOST,
}

NO_CDN_SUPPORT_ERROR = "CDN feature not implemented"


class ScalewayStorageDriver(BaseS3StorageDriver):
    name = "Scaleway Storage Driver"
    website = "https://www.scaleway.com/en/object-storage/"
    connectionCls = S3SignatureV4Connection
    region_name = "fr-par"

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        region="fr-par",
        url=None,
        **kwargs,
    ):
        # Here for backward compatibility for old and deprecated driver class
        # per region approach
        if hasattr(self, "region_name") and not region:
            region = self.region_name  # pylint: disable=no-member

        self.region_name = region

        if region and region not in REGION_TO_HOST_MAP.keys():
            raise ValueError("Invalid or unsupported region: %s" % (region))

        self.name = "Scaleway Object Storage (%s)" % (region)

        if host is None:
            self.connectionCls.host = REGION_TO_HOST_MAP[region]
        else:
            self.connectionCls.host = host

        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            region=region,
            url=url,
            **kwargs,
        )

    @classmethod
    def list_regions(self):
        return REGION_TO_HOST_MAP.keys()

    def get_object_cdn_url(self, obj, ex_expiry=S3_CDN_URL_EXPIRY_HOURS):
        # In order to download (private) objects we need to be able to generate a valid CDN URL,
        # hence shamefully just use the working code from the S3StorageDriver.
        return S3StorageDriver.get_object_cdn_url(self, obj, ex_expiry=ex_expiry)
